#!/usr/bin/env python
# ===================================================================
# Purpose:      Functions for SerialEM utilities needed by other packages and scripts.
# Author:       Fabian Eisenstein
# Created:      2024/08/13
# ===================================================================

# Check SerialEM availability
try:
    import serialem as sem
    # Check if PySEMSocket is connected
    try:
        sem.Delay(0, "s")
        SERIALEM = True
    except sem.SEMmoduleError:
        SERIALEM = False
except ModuleNotFoundError:
    SERIALEM = False

import sys
import time
from pathlib import Path
from SPACEtomo.modules.utils import log
from SPACEtomo import __version__, version_SerialEM

def connectSerialEM(ip="127.0.0.1", port=48888):
    """Establishes connection to SerialEM."""
    
    sem.ConnectToSEM(port, ip)

    sem.SuppressReports()
    sem.ErrorsToLog()
    sem.SetNewFileType(0)           # set file type to mrc in case user changed default file type

def checkVersion():
    """Check for minimal version of SerialEM."""

    version, date_string = version_SerialEM
    versionCheck = sem.IsVersionAtLeast(version, date_string)
    if not versionCheck and sem.IsVariableDefined("warning_version") == 0:
        runScript = sem.YesNoBox("\n".join(["WARNING: You are using a version of SerialEM that does not support all SPACEtomo features. It is recommended to update to the latest SerialEM beta version!", "", "Do you want to attempt to run SPACEtomo regardless?"]))
        if not runScript:
            sem.Exit()
        else:
            sem.SetPersistentVar("warning_version", "")

def getSessionDir():
    """Sets up top level dir for SPACEtomo session."""

    if sem.IsVariableDefined("ses_dir") == 0:
        choice = sem.UserSetDirectory("Please choose a directory for this session!")
        if choice == "Cancel":
            log("ERROR: No session directory was selected. Aborting run!")
            sem.Exit()
        ses_dir = sem.ReportDirectory()
        sem.SetPersistentVar("ses_dir", ses_dir)
    else:
        ses_dir = sem.GetVariable("ses_dir")
        sem.SetDirectory(str(ses_dir))

    return Path(ses_dir)

def getGridList(grid_list, automation_level):
    """Checks which grids need to be processed."""

    if not isinstance(grid_list, list):
        grid_list = [grid_list]

    # Check if this is not first grid in list
    if sem.IsVariableDefined("grid_list") == 1:
        remaining_grid_list = sem.GetVariable("grid_list").split(",")

        # If grid list is empty, finish run
        if len(remaining_grid_list) == 1 and remaining_grid_list[0] == "":
            exitSPACEtomo()

        else:
            # Convert to int
            remaining_grid_list = [int(no) for no in remaining_grid_list]

            # Check if grid list was changed since the persistent var was created
            if not all(grid in grid_list for grid in remaining_grid_list):
                log("WARNING: Grid list was updated and previously saved grid list was discarded!")
                remaining_grid_list = grid_list

    else:
        remaining_grid_list = grid_list

    # Save rest of list in persistent var for later calls of script (only necessary when acquisition is automated)
    if automation_level >= 5:
        sem.SetPersistentVar("grid_list", ",".join(str(no) for no in remaining_grid_list[1:]))
        sem.SetStatusLine(5, "Grid: " + str(len(grid_list) - len(remaining_grid_list) + 1)  + " / " + str(len(grid_list)))

        # Only run all steps on current grid, then run PACEtomo before running SPACEtomo again
        remaining_grid_list = remaining_grid_list[:1]

    return remaining_grid_list

def prepareEnvironment(session_dir, grid_name, external_dir):
    """Sets up dirs and logs."""

    # Save old log before switching dir
    openNewLog()

    # Set subfolder for grid
    cur_dir = session_dir / grid_name
    cur_dir.mkdir(exist_ok=True)
    sem.SetDirectory(str(cur_dir))

    # Opens log in new dir
    openNewLog(grid_name, force=True)

    # Check if external processing directory is valid
    external = False
    if external_dir == "":
        try:
            import ultralytics
        except ModuleNotFoundError:
            log(f"ERROR: Packages for local processing could not be found. Please set up external processing!")
            sem.exit()
        map_dir = cur_dir / "SPACE_maps"
        map_dir.mkdir(exist_ok=True)
    else:
        map_dir = Path(external_dir)
        if map_dir.exists():
            external = True
        else:
            log(f"ERROR: External map [{map_dir}] directory does not exist!")
            sem.Exit()

    # Close all files
    while sem.ReportFileNumber() > 0:
        sem.CloseFile()

    return cur_dir, map_dir, external

def openNewLog(name=None, force=False):
    if not force:
        if name:
            sem.SaveLogOpenNew(name)
        else:
            sem.SaveLogOpenNew()
    else:
        sem.CloseLogOpenNew(1)
        if name:
            sem.SaveLog(1, name)
    
    # Start new log with version numbers
    if name:
        log(f"SPACEtomo Version {__version__}")
        sem.ProgramTimeStamps()
        log(f"DEBUG: Python {sys.version}")

def setupAcquisition(spacetomo_script_id, postaction_script_id, pacetomo_script_id):
    num_acq, *_ = sem.ReportNumNavAcquire()
    log(f"Starting PACEtomo acquisition of {int(num_acq)} areas!")

    # Setting Acquire at Items scripts
    sem.NavAcqAtEndUseParams("F")
    sem.SetNavAcqAtEndParams("prim", 2)
    sem.SetNavAcqAtEndParams("scrp-p", pacetomo_script_id)      # PACEtomo script
    sem.SetNavAcqAtEndParams("scrp-b", 0)
    sem.SetNavAcqAtEndParams("scrp-a", postaction_script_id)    # SPACEtomo_postAction script
    sem.RunScriptAfterNavAcquire(spacetomo_script_id)           # SPACEtomo script

    sem.StartNavAcquireAtEnd()

def confirmationBox(text):
    """Opens a popup box for user confirmation and script cancel option."""

    sem.Pause(text)

def exitSPACEtomo():
    """Successfully exit session."""

    log("##### SPACEtomo run completed! #####")
    log(time.strftime("%d.%m.%Y %H:%M:%S", time.localtime()))
    sem.ClearStatusLine(0)
    sem.ClearPersistentVars()
    sem.Exit()


#!Python
# ===================================================================
#ScriptName SPACEtomo
# Purpose:      Finds lamellae in grid map, collects lamella montages and finds targets using deep learning models.
#               More information at http://github.com/eisfabian/PACEtomo
# Author:       Fabian Eisenstein
# Created:      2023/05/31
# ===================================================================

### TODO before running:
# - Microscope alignments
# - Run inventory and name grids in autoloader
# - Low Dose mode setup (ensure that both WG image state and MM (Low Dose) image states produce good images)
# - Set Shift offset between Record and View mag and check if "High Def IS" calibrations are necessary (Apply large IS in View and check if Record acquires at same position)
# - PACEtomo preparation and script settings (for automation level 5)
# - Setup "Acquire at Items" dialogue to run PACEtomo as main action and SPACEtomo_postAction as "Run after primary action" (for automation level 5)
# - In case of inference on an external GPU workstation, start SPACEtomo_monitor.py script and set external_map_dir accordingly
###

############ SETTINGS ############ 

automation_level        = 4                     # = 1: collect WG map and find lamellae
                                                # = 2: collect MM maps for each lamella
                                                # = 3: segment lamella MM maps
                                                # = 4: setup targets based on segmentation for each lamella
                                                # = 5: start PACEtomo batch acquisition of all targets



script_numbers          = [10, 11, 13]          # Numbers of [SPACEtomo_run.py, SPACEtomo_postAction.py, PACEtomo.py] scripts in SerialEM script editor
external_map_dir        = ""                    # Path to directory where maps are saved and segmentations are expected (run "SPACEmonitor" on external machine to manage runs and queue)

# Grid settings [Level 1]

grid_list               = [1]                   # Number or list of numbers of target grid(s) in autoloader
exclude_lamella_classes = ["gone"]              # Lamella classes to exclude, possible classes: ["broken", "gone", "thick", "wedge", "contaminated", "good"]

WG_image_state          = 1                     # Imaging state name or index for whole grid montage (set up before!)
IM_mag_index            = 10                    # Intermediate mag index to use (10: 580x) 
WG_distance_threshold   = 5                     # Minimum distance [microns] between lamellae to not be considered duplicate detection

WG_wait_for_inspection  = False                 # = True: wait for inspection of lamellae detection by operator using the GUI

# Lamella montage settings [Level 2]

MM_image_state          = [13]                  # Imaging state name or index for Low Dose Mode, this can be several imaging states to specify Record, View, ... (set up before!)

# Targeting settings [Level 4]

manual_selection        = False                 # = True: creates empty segmentation, allows for manual target setup by operator using the GUI
MM_wait_for_inspection  = False                 # = True: wait for inspection of targets using the GUI

target_list             = ["mitos"]             # List of target classes, additive, possible classes: ["background", "white", "black", "crack", "coating", "cell", "cellwall", "nucleus", "vacuole", "mitos", "lipiddroplets", "vesicles", "multivesicles", "membranes", "dynabeads", "ice", "cryst", "lamella"]
avoid_list              = ["black", "white", "ice", "crack", "dynabeads"]    # list of classes to avoid, additive
target_score_threshold  = 0.01                  # Weighted fraction of FOV containing the target, keep at 0 for any structures smaller than the FOV
sparse_targets          = True                  # Use for sparse targets on the sample (e.g. mitos, vesicles)
target_edge             = False                 # Use for edge targets (e.g. nuclear envelope, cell periphery)
penalty_weight          = 0.3                   # Factor to downweight overlap with classes to avoid

extra_tracking          = False                 # Add an extra center target for the tracking tilt series (not working properly yet)
max_tilt                = 60                    # Maximum tilt angle [degrees] of tilt series to be acquired 

########## END SETTINGS ########## 

try:
    import serialem as sem
except ModuleNotFoundError:
    print("WARNING: Trying to run SerialEM externally!")

from SPACEtomo.run import main
import SPACEtomo.modules.utils as utils

cur_dir = utils.getCurDir()
utils.saveSettings(cur_dir / "SPACEtomo_settings.ini", globals())
main()



#!/usr/bin/env python
# ===================================================================
# ScriptName:   nav
# Purpose:      Interface for SerialEM navigator.
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/08/13
# Revision:     v1.2
# Last Change:  2024/09/04: added newPolygon function
#               2024/08/23: allowed creation of minimal map without template
#               2024/08/19: added creation of new map and point items
#               2024/08/16: added change item functions, added writeToFile, newMap, replaceItem functions 
#               2024/08/14: added search function 
# ===================================================================

# Check SerialEM availability
try:
    import serialem as sem
    # Check if PySEMSocket is connected
    try:
        sem.Delay(0, "s")
        SERIALEM = True
    except sem.SEMmoduleError:
        SERIALEM = False
except ModuleNotFoundError:
    SERIALEM = False

from pathlib import Path
import numpy as np
import mrcfile

from SPACEtomo.modules.buf import Buffer
from SPACEtomo.modules.utils import log
from SPACEtomo.modules.utils import serialem_check, dummy_skip, dummy_replace

class NavItem:
    """Nav item."""

    def __init__(self, index, label) -> None:
        self.nav_index = int(index)      # SerialEM navigator item index (starting from 1, SPACEtomo Navigator IDs start from 0)
        self.label = str(label)
        self.entries = {}

        # Instantiate attributes
        self.stage = np.zeros(3)
        self.item_type = 0
        self.note = ""
        self.map_file = None

    def readBlock(self, open_file):
        """Reads block of open navigator file until next empty line."""

        for line in open_file:
            col = line.strip().split(" ")
            if col[0] == "":
                return
            self.entries[col[0]] = [val for val in col[2:]]

    def getBlock(self):
        """Converts attributes back to navigator file entry."""

        # Check existence of map and downgrade to polygon if map file cannot be found to prevent error when loading nav
        if self.item_type == 2 and self.map_file:
            if not self.map_file.exists():
                self.item_type = 1

        # Update changed attributes
        self.revertEntries()

        # Create text block
        text = f"[Item = {self.label}]\n"
        for key, attr in self.entries.items():
            text += f"{key} = {' '.join(attr)}\n"
        text += "\n"

        return text

    def convertEntries(self):
        """Exposes relevant entries as attributes. [StageXYZ, Type, Note, MapFile]"""

        self.stage = np.array(self.entries["StageXYZ"], dtype=np.float32)               # Stage position
        self.item_type = int(self.entries["Type"][0])                                   # 0-2 for point, polygon, map
        if "Note" in self.entries.keys():
            self.note = " ".join(self.entries["Note"])                                  # Note string
        else:
            self.note = ""
        if "MapFile" in self.entries.keys():
            self.map_file = Path(" ".join(self.entries["MapFile"]))                     # Parse map file
        else:
            self.map_file = None

    def revertEntries(self):
        """Updates entries according to changed attributes. [StageXYZ, Type, Note, MapFile]"""

        self.entries["StageXYZ"] = [str(coord) for coord in self.stage]                 # Stage position          
        self.entries["Type"] = [str(self.item_type)]                                    # 0-2 for point, polygon, map
        if self.note:
            self.entries["Note"] = self.note.split(" ")                                 # Note string
        if self.map_file:
            self.entries["MapFile"] = [str(self.map_file)]                              # Parse map file

    def changeColor(self, color_id):
        """Changes color of nav item (0 = red; 1 = green; 2 = blue; 3 = yellow; 4 = magenta; 5 = no realign)."""

        self.entries["Color"] = [str(color_id)]

        if SERIALEM:
            sem.ChangeItemColor(self.nav_index, color_id)

    def changeLabel(self, label):
        """Changes label of nav item."""

        self.label = label

        if SERIALEM:
            sem.ChangeItemLabel(self.nav_index, label)
            
    def changeNote(self, note):
        """Changes note of nav item."""

        self.note = note

        if SERIALEM:
            sem.ChangeItemNote(self.nav_index, note)

    def changeDraw(self, draw):
        """Changes if nav item is drawn."""

        self.entries["Drawn"] = [str(int(draw))]

        if SERIALEM:
            sem.ChangeItemDraw(self.nav_index, int(draw))

    def changeAcquire(self, acquire):
        """Changes acquire flag for item."""

        self.entries["Acquire"] = [str(acquire)]

        if SERIALEM:
            sem.SetItemAcquire(self.nav_index)

    def changeStage(self, coords, relative=False):
        """Changes stage coordinates."""

        # Add z coord if not given
        if len(coords) < 3:
            if relative:
                coords = np.append(coords, 0)
            else:
                coords = np.append(coords, self.stage[2])

        # Get relative pts coords
        pts_x = [float(x) - self.stage[0] for x in self.entries["PtsX"]]
        pts_y = [float(y) - self.stage[1] for y in self.entries["PtsY"]]

        if relative:
            self.stage += coords
        else:
            self.stage = coords

        self.entries["PtsX"] = [str(x + self.stage[0]) for x in pts_x]
        self.entries["PtsY"] = [str(y + self.stage[1]) for y in pts_y]

        self.revertEntries()

    def changeZ(self, z, relative=False):
        """Updates z coordinates of nav item."""

        if relative:
            self.stage[2] += z
        else:
            self.stage[2] = z
        self.revertEntries()

    def addEntry(self, key, value):
        """Adds any entry to nav item."""

        if isinstance(value, list):
            self.entries[key] = [str(item) for item in value]
        else:
            self.entries[key] = [str(value)]

    def getDistance(self, coords):
        """Calculates distance in x and y from given coords."""

        coords = np.array(coords)
        #log(f"DEBUG: Distance: {self.stage[:2]}, {coords}, {np.linalg.norm(self.stage[:2] - coords)}")
        return np.linalg.norm(self.stage[:2] - coords)
    
    def getVector(self, coords):
        """Calculates vector to coords."""

        coords = np.array(coords)
        return coords - self.stage[:2]
    
    def createMap(self, template, stage, file, map_id, note=""):
        """Creates new map item from image file and template."""

        with mrcfile.open(file) as mrc:
            min_max_scale = [str(np.min(mrc.data)), str(np.max(mrc.data))]
            dims = np.flip(mrc.data.shape)

        # Generate default template if no template
        if not template:
            template = self.getMinimalTemplate()

        # Get stage Z from template if missing
        if len(stage) < 3:
            stage = np.array([stage[0], stage[1], template.stage[2]])

        # Calculate point coords using dimensions and inverse of MapScaleMat matrix
        points = np.array([[-0.5, -0.5], [-0.5, 0.5], [0.5, 0.5], [0.5, -0.5], [-0.5, -0.5]]) * dims
        px2stage = np.linalg.inv(np.array(template.entries["MapScaleMat"], dtype=np.float32).reshape((2, 2)))
        points = np.array([px2stage @ point for point in points]) + stage[:2]
        pts_x, pts_y = points[:, 0], points[:, 1]

        # Go through all required entries
        self.entries["Color"] = ["2"]                                           # Default map color is blue
        self.entries["StageXYZ"] = [str(x) for x in stage]                      # XYZ stage coords
        self.entries["NumPts"] = ["5"]                                          # 5 points to define rectangle
        self.entries["Regis"] = template.entries["Regis"]                       # Registration
        self.entries["Type"] = ["2"]                                            # Type map is 2
        self.entries["Note"] = note.split(" ") if isinstance(note, str) else note
        self.entries["SamePosId"] = [str(map_id)]                               # ID of map at same coords
        self.entries["MapFile"] = [str(file)]                                   # Image file of map
        self.entries["MapID"] = [str(map_id)]                                   # Map has to be unique
        self.entries["MapMontage"] = ["0"]                                      # Map is not a montage
        self.entries["MapSection"] = ["0"]                                      # Map is only section in file
        self.entries["MapBinning"] = template.entries["MapBinning"]             # Map binning
        self.entries["MapMagInd"] = template.entries["MapMagInd"]               # Map mag index
        self.entries["MapCamera"] = template.entries["MapCamera"]               # Camera number
        self.entries["MapScaleMat"] = template.entries["MapScaleMat"]           # Stage to pixel scale matrix for drawing (might be off when taken from montage template)
        self.entries["MapWidthHeight"] = [str(dim) for dim in dims]             # Size of initial map image

        # Optional entries
        self.entries["MapMinMaxScale"] = min_max_scale                          # Min and max scale values of image when map was defined
        self.entries["MapFramesXY"] = ["0", "0"]                                # Number of montage frames
        self.entries["ImageType"] = ["0"]                                       # 0 for mrc file

        other_entries = ["MontBinning", "MapExposure", "MapSpotSize", "MapIntensity", "MapSlitIn", "MapSlitWidth", "DefocusOffset", "K2ReadMode", "NetViewShiftXY", "ViewBeamShiftXY", "ViewBeamTiltXY", "MapProbeMode", "MapLDConSet", "MapTiltAngle"]
        for entry in other_entries:
            if entry in template.entries.keys():
                self.entries[entry] = template.entries[entry]

        self.entries["PtsX"] = [str(pt) for pt in pts_x]
        self.entries["PtsY"] = [str(pt) for pt in pts_y]

        self.convertEntries()

    def createPoint(self, coords, map_id, group_id=None, note=""):
        """Creates new point item."""

        self.entries["Color"] = ["0"]
        if len(coords) < 3:
            raise ValueError("Point item needs 3D coords but only 2D were given!")
        self.entries["StageXYZ"] = [str(x) for x in coords]
        self.entries["NumPts"] = ["1"]
        self.entries["Regis"] = ["1"]
        self.entries["Type"] = ["0"]
        if group_id is not None:
            self.entries["GroupID"] = [str(group_id)]
        self.entries["MapID"] = [str(map_id)] 
        self.entries["PtsX"] = [str(coords[0])]
        self.entries["PtsY"] = [str(coords[1])]

        self.entries["Note"] = note.split(" ") if isinstance(note, str) else note

        self.convertEntries()

    def createPolygon(self, pts_x, pts_y, z, map_id, note=""):
        """Creates new point item."""

        center = np.array([np.mean(pts_x), np.mean(pts_y), z])

        self.entries["Color"] = ["1"]
        self.entries["StageXYZ"] = [str(x) for x in center]
        self.entries["NumPts"] = [str(len(pts_x))]
        self.entries["Regis"] = ["1"]
        self.entries["Type"] = ["1"]
        self.entries["MapID"] = [str(map_id)] 
        self.entries["PtsX"] = [str(x) for x in pts_x]
        self.entries["PtsY"] = [str(y) for y in pts_y]

        self.entries["Note"] = note.split(" ") if isinstance(note, str) else note

        self.convertEntries()

    """
    def defaults(self):
        self.color = 0              # Color         (0 = red; 1 = green; 2 = blue; 3 = yellow; 4 = magenta; 5 = no realign)
        self.stage = [0, 0, 0]      # StageXYZ      Stage position
        self.num_pts = 0            # NumPts        Number of points
        self.draw = 1               # Draw          Flag to draw
        self.item_type = 0          # Type          0-2 for point, polygon, map
        self.note = ""              # Note          Note string
        self.group_id = 0           # GroupID       ID of group that item belongs to
        self.imported = 0           # Imported      Indicator of an imported map or point drawn on one
        self.drawn_id = 0           # DrawnID       ID of map point/polygon was drawn on
        self.same_pos_id = 0        # SamePosId     Items with matching IDs were taken at same raw stage position
        self.acquire = 0            # Acquire       Flag for acquiring
        self.map_file = ""          # MapFile       Full or relative path of map file
        self.map_id = 0             # MapID         Unique ID (all items get one, not just maps)
        self.map_montage = 0        # MapMontage    Flag that map is a montage
        self.map_section = 0        # MapSection    Section number in file
        self.map_binning = 1        # MapBinning    Binning at which map was taken, or of initial overview map image for montage
        self.map_mag_id = 0         # MapMagInd     Magnification index of map, or of non-map image a point or polygon was drawn on
        self.grid_map_transform = 

        self.attr_entries = ["Color", "StageXYZ", "NumPts", "Draw", "Type", "Note", ]
        self.other_entries = {}     # Catch all for entries that don't have attribute
    """
    @staticmethod
    def getMinimalTemplate():
        template = NavItem(index=0, label="template")
        template.stage = np.zeros(3)

        template.entries["Regis"] = ["1"]                                       # Registration
        template.entries["MapBinning"] = ["1"]                                  # Map binning
        template.entries["MapMagInd"] = ["0"]                                   # Map mag index
        template.entries["MapCamera"] = ["1"]                                   # Camera number
        template.entries["MapScaleMat"] = ["1.0", "0.0", "0.0", "1.0"]          # Stage to pixel scale matrix for drawing (might be off when taken from montage template)

        return template

class Navigator:

    map_id_counter = 11

    def __init__(self, file=None, is_open=False) -> None:
        self.file = file
        self.header = ""
        self.items = []

        if self.file:
            self.file = Path(file)
            if file.exists():
                self.loadOld()
                self.readFromFile()
            else:
                self.openNew()
        elif is_open:
            self.getOpen()

    def readFromFile(self):
        """Reads lines from navigator text file and delegates to NavItem to read item blocks."""

        with open(self.file) as f:
            for line in f:
                col = line.strip().split(" ")
                if col[0] == "": continue
                if col[0] == "[Item":
                    label = col[2].strip("]")
                    self.items.append(NavItem(index=len(self.items) + 1, label=label))
                    self.items[-1].readBlock(f)
                    self.items[-1].convertEntries()
                else:
                    self.header += line

    def writeToFile(self, file=None):
        """Write new navigator file."""

        if file:
            file = Path(file)
        else:
            file = self.file
        if not file:
            raise ValueError("File for saving navigator could not be identified!")

        # Make sure header contains at least adoc version
        if not self.header:
            self.header = "AdocVersion = 2.00\n"

        text = self.header + "\n"

        for item in self.items:
            text += item.getBlock()

        with open(file, "w") as f:
            f.write(text)

    @dummy_skip
    @serialem_check
    def loadOld(self):
        """Load nav file."""

        nav_status = sem.ReportIfNavOpen()
        # If a nav file is open
        if nav_status == 2:
            sem.SaveNavigator()

            # Check if current nav file is from loaded grid
            nav_file = Path(sem.ReportNavFile())
            if nav_file == self.file:
                return

        # If nav is not saved
        elif nav_status == 1:
            sem.SaveNavigator("temp.nav")
            log("WARNING: Open navigator was saved as temp.nav and closed!")

        # If nav is not open
        else:
            sem.OpenNavigator()

        # Load nav file
        sem.ReadNavFile(str(self.file))

    @dummy_skip
    @serialem_check
    def openNew(self):
        """Open new nav file."""

        nav_status = sem.ReportIfNavOpen()
        # If a nav file is open
        if nav_status == 2:
            sem.SaveNavigator()
            sem.CloseNavigator()

        # If nav is not saved
        elif nav_status == 1:
            sem.SaveNavigator("temp.nav")
            sem.CloseNavigator()
            log("WARNING: Open navigator was saved as temp.nav and closed!")

        # Open new nav
        sem.OpenNavigator(str(self.file))

    @dummy_skip
    @serialem_check
    def getOpen(self):
        """Gets file from open nav in SerialEM."""

        self.file = Path(sem.ReportNavFile())
        self.pull()

    def pull(self):
        """Reloads nav file pulling changes from SerialEM."""

        # If SerialEM is open, save navigator
        if SERIALEM:
            sem.SaveNavigator(str(self.file))
        
        # Reset attributes and reload
        self.header = ""
        self.items = []
        self.readFromFile()

    def push(self):
        """Reloads nav file pushing changes in instance to SerialEM."""

        # If SerialEM is open, save navigator
        if SERIALEM:
            sem.SaveNavigator(str(self.file))
            sem.CloseNavigator()
        
        # Reset attributes and reload
        self.writeToFile()

        if SERIALEM:
            sem.ReadNavFile(str(self.file))

    @dummy_skip
    @serialem_check
    def shiftItems(self, shift):
        """Shift all navigator items by shift in microns."""
        
        log(f"Shifting items by {shift}!")
        sem.ShiftItemsByMicrons(*shift)

        self.pull()

    @dummy_replace("newMapFromImg")
    @serialem_check
    def newMap(self, buffer=None, img_file="", label="", note="", **kwargs):                                               # Only accept keyword args to make compatible with newMapFromImg
        """Makes new navigator map from buffer."""
        
        if not isinstance(buffer, Buffer):
            raise ValueError("Map has to be a buffer object!")
        
        img_file = Path(img_file)

        # Load buffer
        buffer.show()

        # Load or create file
        if img_file.exists():
            if not sem.IsImageFileOpen(str(img_file)):
                sem.OpenOldFile(str(img_file))
        else:
            sem.OpenNewFile(str(img_file))
            sem.S()

        # Make map
        nav_id = int(sem.NewMap()) - 1

        if label:
            sem.ChangeItemLabel(nav_id + 1, label)
        if note:
            sem.ChangeItemNote(nav_id + 1, note)

        # Close file
        sem.CloseFile()

        # Add nav_id to buffer
        buffer.nav_id = nav_id

        # Reload nav
        self.pull()

        return nav_id
    
    def newMapFromImg(self, img_file="", template_id=None, coords=(0, 0, 0), label="", note="", **kwargs):      # Only accept keyword args to make compatible with newMap
        """Makes new map from image file without SerialEM."""

        if template_id is not None:     # If template is None, a minimal template with default values will be used
            template = self.items[template_id]
        else:
            template = None

        nav_item = NavItem(len(self), label=label)
        nav_item.createMap(template, coords, img_file, self.map_id_counter, note)
        self.map_id_counter += 1

        self.items.append(nav_item)

        # Reload nav
        self.push()

        return len(self) - 1
    
    def newPoint(self, coords, label="", note="", color_id=0, group_id=None, update=True):
        """Adds new point to navigator."""

        if not label:
            label = str(len(self))

        nav_item = NavItem(len(self), label=label)
        nav_item.createPoint(coords, self.map_id_counter, group_id=group_id, note=note)
        self.map_id_counter += 1
        nav_item.entries["Color"] = [str(color_id)]

        self.items.append(nav_item)

        # Reload nav
        if update:
            self.push()

        return len(self) - 1    # nav_id
    
    def newPointGroup(self, points, label_prefix, color_id=0, stage_z=0):
        """Adds group of points to navigator."""

        group_id = self.map_id_counter
        self.map_id_counter += 1

        for p, point in enumerate(points):
            label = label_prefix + str(p + 1).zfill(len(str(len(points))))
            coords = np.array([point[0], point[1], stage_z])
            nav_id = self.newPoint(coords, label, label, color_id, group_id, update=False)

        # Reload nav
        self.push()

    def newPolygon(self, pts_x, pts_y, z, label="", note="", color=3, update=True):
        """Adds polygon to navigator."""

        if not label:
            label = str(len(self))

        nav_item = NavItem(len(self), label=label)
        nav_item.createPolygon(pts_x, pts_y, z, map_id=self.map_id_counter, note=note)
        self.map_id_counter += 1

        nav_item.entries["Color"] = [str(color)]

        self.items.append(nav_item)

        # Reload nav
        if update:
            self.push()

        return len(self) - 1    # nav_id

    def replaceItem(self, old_id, new_id):
        """Replaces an old item with a new item."""

        self.items[old_id] = self.items[new_id]
        self.items.pop(new_id)

        # Reload nav
        self.push()

    def getIDfromNote(self, note, warn=True):
        """Finds index if item by note."""

        candidates = [id for id, item in enumerate(self.items) if item.note == note]

        if not candidates:
            if warn:
                log(f"WARNING: No navigator item has the note [{note}]!")
            return None
            #raise LookupError(f"No navigator item found with note [{note}]!")

        if len(candidates) > 1:
            if warn:
                log(f"WARNING: More than one navigator items has the note [{note}]!")

        return candidates[0]
    
    def searchByEntry(self, entry, query, partial=False, subset=[]):
        """Search navigator items.

        Args:
            entry: Name of navigator item entry.
            query: Search query for specified entry.
            partial: Return items where query in entry instead of query == entry.
            subset: Only include if item ID is in this list.
        """

        results = []
        for id, item in enumerate(self.items):
            # Exclude items not in subset
            if subset and id not in subset: continue

            # Deal with special item entries
            if entry == "label":
                value = item.label
            elif entry == "stage":
                value = item.stage
            else:
                if entry in item.entries.keys():
                    # Concatenate entry back to string
                    value = " ".join(item.entries[entry])
                else:
                    continue
            
            # Check for match
            if not partial:
                if value == query:
                    results.append(id)
            else:
                if query in value:
                    results.append(id)

        return results
    
    def searchByCoords(self, coords, margin=0, subset=None):
        """Search navigator stage coords in X and Y.

        Args:
            coords: Coords at which to search.
            margin: Accepted distance from given coords.
            subset: Only include if item ID is in this list.
        """

        coords = np.array(coords, dtype=np.float32)

        results = []
        for id, item in enumerate(self.items):
            # Exclude items not in subset
            if subset is not None and id not in subset: continue

            # Calculate distance
            dist = item.getDistance(coords)

            # Check for match
            if dist <= margin:
                results.append(id)

        return results

    def __len__(self):
        return len(self.items)
    
    def __iter__(self):
        return iter(self.items)








if __name__ == "__main__":
    test_nav = Path("/Users/eifabian/Documents/temp/GW282_g4.nav")

    nav = Navigator(test_nav)

    print(len(nav))
    print(nav.header)

    print(nav.getIDfromNote("good (86%)"))

    print(nav.getIDfromNote("20240528_GW282_g4_l10_tgts.txt"))

    #print(nav.getIDfromNote("fail"))

    result = nav.searchByEntry("label", "PL", partial=True)
    print(result)

    coords = (318.903, -699.347)
    result = nav.searchByCoords(coords, 5, result)
    print(result)

    for item in result:
        print(nav.items[item].label, nav.items[item].stage)

    nav.writeToFile(test_nav.parent / (test_nav.stem + "_rewritten.nav"))

    print(nav.items[5].entries["Color"])


#!/usr/bin/env python
# ===================================================================
# ScriptName:   gui_menu
# Purpose:      User interface utility to generate menus
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/08/08
# Revision:     v1.2
# Last Change:  2024/10/02: added slider support
#               2024/08/08: separated from gui.py 
# ===================================================================

import dearpygui.dearpygui as dpg

class Menu:
    """Class to structure menu tables and control locking and unlocking of rows."""

    def __init__(self) -> None:
        self.makeTable()
        self.rows = {}                  # visible rows
        self.rows_locked = {}           # conditionally hidden rows
        self.rows_adv = {}              # advanced rows
        self.rows_adv_locked = {}       # advanced conditionally hidden rows
        self.elements = {}              # elements within rows
        self.elements_adv = {}          # advanced elements
        self.separators = {}            # separators between rows
        self.last_group = None          # group to which new elements will be added

        self.show_advanced = False      # toggle advance elements/rows

    def makeTable(self):
        """Creates base table."""

        with dpg.table(header_row=False, width=-1, borders_innerH=True) as self.table:
            dpg.add_table_column()

    def newRow(self, tag, horizontal=False, separator=False, locked=True, advanced=False):
        """Creates new row and necessary containers for elements."""

        with dpg.table_row(parent=self.table):
            with dpg.table_cell():
                self.last_group = dpg.add_group(horizontal=horizontal, show=not locked and not advanced)
                if separator:
                    self.separators[tag] = dpg.add_separator(show=not locked and not advanced)
        if advanced and locked:
            self.rows_adv_locked[tag] = self.last_group
        elif advanced:
            self.rows_adv[tag] = self.last_group
        elif locked:
            self.rows_locked[tag] = self.last_group
        else:
            self.rows[tag] = self.last_group

    def unlockRows(self, tags=[]):
        """Makes list of rows visible."""

        for tag in tags:
            # Normal rows
            if tag in self.rows_locked.keys():
                self.rows[tag] = self.rows_locked[tag]
                self.rows_locked.pop(tag)
                dpg.show_item(self.rows[tag])
                # Check for accompanying row separator
                if tag in self.separators:
                    dpg.show_item(self.separators[tag])
            # Advanced rows
            if tag in self.rows_adv_locked.keys():
                self.rows_adv[tag] = self.rows_adv_locked[tag]
                self.rows_adv_locked.pop(tag)
                # Only show in advanced mode
                if self.show_advanced:
                    dpg.show_item(self.rows_adv[tag])
                    # Check for accompanying row separator
                    if tag in self.separators:
                        dpg.show_item(self.separators[tag])

    def lockRows(self, tags=[]):
        """Hides list of rows until unlocked."""

        for tag in tags:
            # Normal rows
            if tag in self.rows.keys():
                self.rows_locked[tag] = self.rows[tag]
                self.rows.pop(tag)
                dpg.hide_item(self.rows_locked[tag])
                # Check for accompanying row separator
                if tag in self.separators:
                    dpg.hide_item(self.separators[tag])
            # Advanced rows
            if tag in self.rows_adv.keys():
                self.rows_adv_locked[tag] = self.rows_adv[tag]
                self.rows_adv.pop(tag)
                dpg.hide_item(self.rows_adv_locked[tag])
                # Check for accompanying row separator
                if tag in self.separators:
                    dpg.hide_item(self.separators[tag])

    def addText(self, tag, value="", color=(255, 255, 255, 255), advanced=False):
        """Adds a text element to the current row."""

        element = dpg.add_text(default_value=value, color=color, show=not advanced, parent=self.last_group)
        # Add to element list
        if advanced:
            self.elements_adv[tag] = element
        else:
            self.elements[tag] = element

    def addButton(self, tag, label="", callback=None, user_data=None, theme=None, show=True, advanced=False):
        """Adds a button to the current row."""

        # Add button to group
        element = dpg.add_button(label=label, callback=callback, user_data=user_data, show=show and not advanced, parent=self.last_group)
        # Add to element list
        if advanced:
            self.elements_adv[tag] = element
        else:
            self.elements[tag] = element

        if theme and dpg.does_item_exist(theme):
            dpg.bind_item_theme(element, theme)

    def addInput(self, tag, label="", value=None, width=50, callback=None, advanced=False):
        """Adds an input field to the current row and uses type of value."""

        # Add input to group
        if isinstance(value, float):
            element = dpg.add_input_float(default_value=value, step=0, width=width, label=label, format="%.3f", callback=callback, show=not advanced, parent=self.last_group)
        elif isinstance(value, int):
            element = dpg.add_input_int(default_value=value, step=0, width=width, label=label, callback=callback, show=not advanced, parent=self.last_group)
        elif isinstance(value, str):
            element = dpg.add_input_text(default_value=value, width=width, label=label, callback=callback, show=not advanced, parent=self.last_group)
        else:
            raise ValueError("No input element available for this type!")
        # Add to element list
        if advanced:
            self.elements_adv[tag] = element
        else:
            self.elements[tag] = element

    def addCheckbox(self, tag, label="", value=False, callback=None, advanced=False):
        """Adds checkbox to current row."""

        # Add checkbox to group
        element = dpg.add_checkbox(label=label, default_value=value, callback=callback, show=not advanced, parent=self.last_group)
        # Add to element list
        if advanced:
            self.elements_adv[tag] = element
        else:
            self.elements[tag] = element

    def addCombo(self, tag, label="", combo_list=[], value="", callback=None, user_data=None, width=0, advanced=False):
        """Adds dropdown menu to current row."""

        # Check if list and value are valid
        if not combo_list:
            combo_list = [""]
        if not value or value not in combo_list:
            value = combo_list[0]
        # Add combo list to group
        element = dpg.add_combo(combo_list, default_value=value, label=label, callback=callback, user_data=user_data, width=width, show=not advanced, parent=self.last_group)
        # Add to element list
        if advanced:
            self.elements_adv[tag] = element
        else:
            self.elements[tag] = element

    def addSlider(self, tag, label="", value=0, value_range=[0, 1], width=50, callback=None, advanced=False):
        """Adds a slider to current row."""

        # Add slider to group
        if isinstance(value, float):
            element = dpg.add_slider_float(default_value=value, min_value=value_range[0], max_value=value_range[1], width=width, label=label, callback=callback, show=not advanced, parent=self.last_group)
        elif isinstance(value, int):
            element = dpg.add_slider_int(default_value=value, min_value=value_range[0], max_value=value_range[1], width=width, label=label, callback=callback, show=not advanced, parent=self.last_group)
        else:
            raise ValueError("No slider available for this type!")
        # Add to element list
        if advanced:
            self.elements_adv[tag] = element
        else:
            self.elements[tag] = element

    @property
    def all_rows(self):
        """Provides list of all rows."""

        return self.rows | self.rows_locked | self.rows_adv | self.rows_adv_locked

    @property
    def all_elements(self):
        """Provides list of all elements."""

        return self.elements | self.elements_adv

    def show(self):
        dpg.show_item(self.table)

    def hide(self):
        dpg.hide_item(self.table)

    def toggleAdvanced(self):
        """Shows or hides rows and elements with the advanced flag."""

        self.show_advanced = not self.show_advanced

        if self.show_advanced:
            for tag, row in self.rows_adv.items():
                dpg.show_item(row)
                if tag in self.separators:
                    dpg.show_item(self.separators[tag])
            for element in self.elements_adv.values():
                dpg.show_item(element)
        else:
            for tag, row in self.rows_adv.items():
                dpg.hide_item(row)
                if tag in self.separators:
                    dpg.hide_item(self.separators[tag])
            for element in self.elements_adv.values():
                dpg.hide_item(element)


#!/usr/bin/env python
# ===================================================================
# Purpose:      Dummy functions for SerialEM utilities needed by other packages and scripts.
# Author:       Fabian Eisenstein
# Created:      2024/08/26
# ===================================================================

import sys
import time
from pathlib import Path
from SPACEtomo.modules.utils import log
from SPACEtomo import __version__, version_SerialEM

def connectSerialEM(ip="127.0.0.1", port=48888):
    """Establishes connection to SerialEM."""
    
    log(f"#DUMMY: Connected to SerialEM IP: {ip}, Port: {port}")

def checkVersion():
    """Check for minimal version of SerialEM."""

    log(f"#DUMMY: Checked SerialEM version.")

def getSessionDir():
    """Sets up top level dir for SPACEtomo session."""

    return Path.cwd()

def getGridList(grid_list, automation_level):
    """Checks which grids need to be processed."""

    if not isinstance(grid_list, list):
        grid_list = [grid_list]

    remaining_grid_list = grid_list

    # Save rest of list in persistent var for later calls of script (only necessary when acquisition is automated)
    if automation_level >= 5:
        # Only run all steps on current grid, then run PACEtomo before running SPACEtomo again
        remaining_grid_list = remaining_grid_list[:1]

    return remaining_grid_list

def prepareEnvironment(session_dir, grid_name, external_dir):
    """Sets up dirs and logs."""

    # Set subfolder for grid
    cur_dir = session_dir / grid_name
    cur_dir.mkdir(exist_ok=True)
    
    # Check if external processing directory is valid
    external = False
    if external_dir == "":
        map_dir = cur_dir / "SPACE_maps"
        map_dir.mkdir(exist_ok=True)
    else:
        map_dir = Path(external_dir)
        if map_dir.exists():
            external = True
        else:
            log(f"ERROR: External map [{map_dir}] directory does not exist!")
            sys.Exit()

    return cur_dir, map_dir, external


def openNewLog(name=None, force=False):
   
    # Start new log with version numbers
    if name:
        log(f"SPACEtomo Version {__version__}")
        log(f"{time.strftime('%Y-%m-%d %H:%M:%S', time.localtime())}")

def setupAcquisition(spacetomo_script_id, postaction_script_id, pacetomo_script_id):
    log(f"Starting PACEtomo acquisition of X areas!")

def confirmationBox(text):
    """Opens a popup box for user confirmation and script cancel option."""

    log(text)

def exitSPACEtomo():
    """Successfully exit session."""

    log("##### SPACEtomo run completed! #####")
    log(time.strftime("%d.%m.%Y %H:%M:%S", time.localtime()))



#!Python
# ===================================================================
#ScriptName SPACEtomo_lamellaDetectionGUI
# Purpose:      Script to start SPACEtomo lamella detection GUI.
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# ===================================================================

import serialem as sem
import sys
from SPACEtomo.modules.utils import log
try:
    import dearpygui.dearpygui as dpg
except:
    log("ERROR: DearPyGUI module not installed! If you cannot install it, please run the target selection GUI from an external machine.")
    sys.exit()

from SPACEtomo.modules.gui.lam_sel import LamellaGUI

def main():
    dpg.create_context()
    main = LamellaGUI()
    main.show()
    dpg.destroy_context()

if __name__ == "__main__":
    main()


#!/usr/bin/env python
# ===================================================================
# ScriptName:   scope
# Purpose:      Interface for microscope functions (inspired by Jonathan Bouvette).
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/08/13
# Revision:     v1.2
# Last Change:  2024/09/04: fixed autoloader check, fixed delay during imaging state change
#               2024/09/02: moved MicroscopeDummy to dummy modules
#               2024/08/27: added generation of detailed MM montage, made MicroscopeDummy work
#               2024/08/23: added MicroscopeDummy
#               2024/08/16: added save option to record, added polygon montage
# ===================================================================

# Check SerialEM availability
try:
    import serialem as sem
    # Check if PySEMSocket is connected
    try:
        sem.Delay(0, "s")
        SERIALEM = True
    except sem.SEMmoduleError:
        SERIALEM = False
except ModuleNotFoundError:
    SERIALEM = False

import sys
import time
import json
from pathlib import Path
import numpy as np

from SPACEtomo.modules import utils
from SPACEtomo.modules.utils import log, serialem_check, dummy_skip

class Microscope:

    def __init__(self, imaging_params) -> None:

        self.imaging_params = imaging_params

        self.low_dose = bool(sem.ReportLowDose()[0])
        self.low_dose_area = None

        # Limits
        self.tilt_limit = sem.ReportProperty("MaximumTiltAngle")
        self.is_limit = sem.ReportProperty("ImageShiftLimit")
        #self.stage_limit = sem.ReportProperty("StageLimits")       # not accessible by script command

        self.imaging_params.IS_limit = self.is_limit

        # Properties
        self.ta_offset = sem.ReportProperty("TiltAxisOffset")
        self.use_coma_vs_is = False

        # Set user settings
        sem.SetUserSetting("DriftProtection", 1)
        sem.SetUserSetting("ShiftToTiltAxis", 1)

        # Check autoloader
        self.loaded_grid = None
        self.autoloader = {}
        self.checkAutoloader()

    ### Properties

    @property
    def stage(self):
        return np.array(sem.ReportStageXYZ())
    
    @property
    def tilt_angle(self):
        return sem.ReportTiltAngle()
    
    @property
    def image_shift(self):
        isx, isy, *_ = sem.ReportImageShift()
        return np.array((isx, isy))
    
    @property
    def speciment_shift(self):
        return np.array(sem.ReportSpecimenShift())
    
    @property
    def defocus(self):
        return float(sem.ReportDefocus())
    
    @property
    def spot_size(self):
        return int(sem.ReportSpotSize())
    
    @property
    def condenser_lens(self):
        return float(sem.ReportPercentC2()[0]) # Returns fractional lens strength as second value
    
    @property
    def beam_tilt(self):
        return np.array(sem.ReportBeamTilt())
    
    @property
    def micro_probe(self):
        return int(sem.ReportProbeMode())
    
    @property
    def magnification(self):
        return int(sem.ReportMag()[0]) # Returns low mag mode as second value, and TEM, EFTEM, STEM as third value
    
    @property
    def camera(self):
        try:
            area = self.low_dose_area if self.low_dose_area else "R"
            read_mode = int(sem.ReportReadMode(area))
        except sem.SEMmoduleError:
            read_mode = None

        return int(sem.CameraProperties()[5]), read_mode
    
    @property
    def energy_filter(self):
        slit_width, energy_loss, slit_in = sem.ReportEnergyFilter()
        return int(slit_in), slit_width, energy_loss
    
    def getMatrices(self, mag_index=0, camera=None):
        matrices = {}
        matrices["s2ss"] = np.array(sem.StageToSpecimenMatrix(mag_index)).reshape((2, 2))
        matrices["ss2s"] = np.array(sem.SpecimenToStageMatrix(mag_index)).reshape((2, 2))
        matrices["is2ss"] = np.array(sem.ISToSpecimenMatrix(mag_index)).reshape((2, 2))
        matrices["ss2is"] = np.array(sem.SpecimenToISMatrix(mag_index)).reshape((2, 2))
        matrices["c2ss"] = np.array(sem.CameraToSpecimenMatrix(mag_index)).reshape((2, 2))
        matrices["ss2c"] = np.array(sem.SpecimenToCameraMatrix(mag_index)).reshape((2, 2))
        return matrices
        
    ### Setters

    def setImageShift(self, image_shift, relative=False):
        """Applies image shift."""

        if not relative:
            sem.SetImageShift(image_shift[0], image_shift[1], 0, int(self.use_coma_vs_is))
        else:
            sem.ImageShiftByUnits(image_shift[0], image_shift[1], 0, int(self.use_coma_vs_is))

    def setSpecimenShift(self, specimen_shift, relative=False):
        """Applies image shift in specimen coordinates."""

        if not relative:
            if self.low_dose_area == "R":
                try:
                    image_shift = self.imaging_params.rec_ss2is_matrix @ specimen_shift
                except (AttributeError, ValueError):
                    log(f"ERROR: Could not convert specimen shift to image shift.")
                    return
            elif self.low_dose_area == "V":
                try:
                    image_shift = self.imaging_params.view_ss2is_matrix @ specimen_shift
                except (AttributeError, ValueError):
                    log(f"ERROR: Could not convert specimen shift to image shift.")
                    return
            sem.SetImageShift(image_shift[0], image_shift[1], 0, int(self.use_coma_vs_is))
        else:
            sem.ImageShiftByMicrons(specimen_shift[0], specimen_shift[1], 0, int(self.use_coma_vs_is))

    def setDefocus(self, defocus, relative=False):
        """Applies defocus."""

        if not relative:
            sem.SetDefocus(defocus)
        else:
            sem.ChangeFocus(defocus)

    def setSpotSize(self, spot_size: int):
        """Sets spot size."""

        sem.SetSpotSize(spot_size)

    def setCondenserLens(self, value):
        """Sets condenser lens/illuminated area."""

        sem.SetPercentC2(value)

    def setBeamTilt(self, tilt_xy):
        """Sets beam tilt."""

        sem.SetBeamTilt(*tilt_xy)

    def setMicroProbe(self, probe_mode: int):
        """Turns micro-probe on (1) or off (0)."""

        if probe_mode != self.micro_probe:
            sem.SetProbeMode(str(probe_mode))

    def setMag(self, mag: int):
        """Switches mag by number or index."""

        # Switch to IM mag depending on if mag index or mag value was given
        if mag < 50:
            sem.SetMagIndex(mag)
        else:
            sem.SetMag(mag)

    def setCamera(self, camera_id: int, read_mode: int= None):
        """Selects camera."""

        sem.SelectCamera(camera_id)
        if read_mode is not None:
            area = self.low_dose_area if self.low_dose_area else "R"
            sem.SetK2ReadMode(area, read_mode)

    def setEnergyFilter(self, slit_in, slit_width, energy_loss=0):
        """Sets energy filter slit settings."""

        sem.SetSlitIn(slit_in)
        sem.SetSlitWidth(slit_width)
        sem.SetEnergyLoss(energy_loss)

    def moveStage(self, xyz):
        """Moves stage to x, y, z."""
        
        sem.MoveStageTo(*xyz)

    def tiltStage(self, tilt, backlash=True):
        """Tilts stage to target angle."""

        # Check tilt limit
        if abs(tilt) > self.tilt_limit:
            log(f"ERROR: {tilt} goes beyond maximum stage tilt [{self.tilt_limit}]!")
            return False

        # Do backlash correction
        if backlash and tilt < self.tilt_angle:
            if abs(tilt - 2) > self.tilt_limit:
                log(f"WARNING: Backlash step was skipped to remain within tilt limits.")
            else:
                sem.TiltTo(tilt - 2)
        
        # Tilt to angle
        sem.TiltTo(tilt)

    ### Simple actions

    def record(self, save=""):
        """Records an image and optionally saves it to file."""

        # Find file to save to
        if save:
            save = Path(save)
            # Check if file is open
            if sem.IsImageFileOpen(str(save)):
                # Search for file number and switch to matching file
                for f in range(1, sem.ReportNumOpenFiles() + 1):
                    if sem.ReportOpenImageFile(f) == save:
                        sem.SwitchToFile(f)
                        break
            else:
                # Open file or create new file
                if save.exists():
                    sem.OpenOldFile(str(save))
                else:
                    sem.OpenNewFile(str(save))

        sem.R()
        if save:
            sem.S()

    def autofocus(self, target=0.0, measure=False):
        """Uses beam tilt routine to determine defocus and optionally adjusts it to target defocus."""

        if not measure:
            sem.SetTargetDefocus(target)
            sem.G(1, -1)
        else:
            sem.G(-2, -1)
        defocus, *_ = sem.ReportAutoFocus()
        drift = np.linalg.norm(sem.ReportFocusDrift())
        return defocus, drift

    def eucentricity(self, rough=True):
        """Call eucentricity routine."""

        if rough:
            sem.Eucentricity(1)         # Rough
        else:
            sem.Eucentricity(6)         # Refine and realign

    def changeImagingState(self, target_state, low_dose_expected=None):
        """Changes SerialEM imaging state."""

        if self.low_dose or self.low_dose_area is None:               # Exit Low Dose and wait to avoid occasional DM crash when switching from LD to LM
            sem.SetLowDoseMode(0)
            sem.SetSlitIn(0)
            time.sleep(1)

        # Deal with list of imaging states for different low dose areas
        if not isinstance(target_state, list):
            target_state = [target_state]

        for state in target_state:
            err = sem.GoToImagingState(str(state))
            if err > 0:
                if err == 1 or err == 3:
                    log(f"ERROR: Imaging state [{state}] not found!")
                elif err == 2:
                    log(f"ERROR: Imaging state [{state}] ambigous!")
                elif err >= 4:
                    log(f"ERROR: Imaging state [{state}] could not be set!")

        # Check if imaging state is a low dose state.
        low_dose, area = sem.ReportLowDose()
        self.low_dose = bool(low_dose)
        self.low_dose_area = area

        # Check if imaging state fits expectation
        if low_dose_expected is not None:
            if self.low_dose != low_dose_expected:
                log(f"WARNING: Imaging state does not fit expected low dose state!")
                return False
        return True

    def changeLowDoseArea(self, area):
        """Changes Low Dose area."""

        if self.low_dose:
            sem.GoToLowDoseArea(area)
        else:
            log("ERROR: Imaging state is not in Low Dose!")

    def changeObjAperture(self, size=0):
        """Inserts objective aperture of given size or retracts aperture if size is 0."""

        try:
            sem.SetApertureSize(2, size)
        except:
            log(f"WARNING: Could not insert objective aperture. Please check if SerialEM has aperture control!")

    def openValves(open=True):
        if open:
            # Open column valves
            sem.SetColumnOrGunValve(1)
        else:
            # Close column valves
            sem.SetColumnOrGunValve(0)

    def checkAutoloader(self):
        """Checks autoloader slots."""

        for slot in range(1, 13):
            slot_status = sem.ReportSlotStatus(slot)
            log(f"DEBUG: Slot status: {slot_status}")
            if slot_status[0] > 0:
                if len(slot_status) > 1 and isinstance(slot_status[-1], str) and slot_status[-1] != "!NONAME!":
                    self.autoloader[slot] = slot_status[-1]
                else:
                    self.autoloader[slot] = "G" + str(slot).zfill(2)
            else:
                if len(slot_status) > 1 and isinstance(slot_status[-1], str) and slot_status[-1] != "!NONAME!":
                    log(f"DEBUG: Slot label: {slot_status[-1]}, Type: {type(slot_status[-1])}, Len: {len(slot_status[-1])}")
                    self.autoloader[slot] = slot_status[-1]
                    if not self.loaded_grid:
                        on_stage = sem.YesNoBox("\n".join(["GRID LOADED?", "", f"Is the grid from slot {slot} [{self.autoloader[slot]}] currently on the stage?"]))
                        if on_stage:
                            self.loaded_grid = slot
                        else:
                            log(f"WARNING: Please only enter names for occupied slots in the Autoloader panel!")

    def loadGrid(self, grid_slot):
        # Initiate loading
        if grid_slot != self.loaded_grid:
            log(f"Loading grid [{self.autoloader[grid_slot]}] from slot {grid_slot}...")
            sem.LoadCartridge(grid_slot)
            self.loaded_grid = grid_slot
        else:
            log(f"Grid [{self.autoloader[grid_slot]}] is already loaded.")

    ### Complex actions

    def collectFullMontage(self, model, overlap):
        """Collects full grid montage."""
        
        grid_name = self.autoloader[self.loaded_grid]

        self.changeImagingState(self.imaging_params.WG_image_state, low_dose_expected=False)
        if self.low_dose:
            log("WARNING: Whole grid montage image state is in Low Dose! This might result in unreasonably large grid maps.")

        # Figure out appropriate binning to get close to model pix size
        pix_size = sem.CameraProperties()[4]
        binning = 1
        while model.pix_size / (pix_size * binning) > 2:
            binning *= 2
        if binning > 2:
            log(f"WARNING: Whole grid montage could be binned by {binning}. Consider using a smaller mag to safe time in the future!")
        #sem.NoMessageBoxOnError(1)
        #sem.SetBinning("R", binning)
        binning = sem.ReportBinning("R") # Temporarily disable setting of binning due to error message in Log

        # Save pix size
        self.imaging_params.WG_pix_size = pix_size * binning

        # Setup montage
        sem.SetupFullMontage(overlap, grid_name + ".mrc")
        #sem.NoMessageBoxOnError(0) # might have to be moved lower down

        # Make sure stage is ready (sometimes stage is busy error popped up when starting montage)
        while sem.ReportStageBusy():
            sem.Delay(1, "s")

        # Collect montage
        log(f"Collecting low mag whole grid montage...")
        sem.M()
        log(f"Making new map from montage...")
        map_id = int(sem.NewMap(0, grid_name + ".mrc")) - 1
        sem.CloseFile()

        sem.ChangeItemLabel(map_id + 1, grid_name[:6])
        if len(grid_name) > 6:
            log(f"WARNING: Grid name [{grid_name}] will be truncated [{grid_name[:6]}] for labels in the Navigator.")

        return map_id

    def collectPolygonMontage(self, poly_id, file, overlap):
        """Collects montage at polygon using View mode."""

        self.changeLowDoseArea("V")
        # Setup polygon montage
        sem.ParamSetToUseForMontage(2)
        sem.SetMontPanelParams(1, 1, 1, 1)      # check all Montage control panel options
        sem.SetupPolygonMontage(poly_id + 1, 0, str(file))
        sem.SetMontageParams(1, int(overlap * min(self.imaging_params.cam_dims)), int(overlap * min(self.imaging_params.cam_dims)), *self.imaging_params.cam_dims, 0, 1) # stage move, overlap X, overlap Y, piece size X, piece size Y, skip correlation, binning

        pieces = int(sem.ReportNumMontagePieces())
        log(f"Collecting medium mag montage at [{file.stem}] ({pieces} pieces)...")

        sem.M()
        map_id = int(sem.NewMap(0, file.name)) - 1
        sem.CloseFile()

        return map_id


class ImagingParams:
    """Previously MicParams."""
    
    file_name = "mic_params.json"

    def __init__(self, WG_image_state, IM_mag_index, MM_image_state, file_dir=None):

        # User settings
        self.WG_image_state = WG_image_state
        self.IM_mag_index = IM_mag_index
        self.MM_image_state = MM_image_state

        # Initialize all attributes
        self.IS_limit = None
        self.cam_dims = np.zeros(2, dtype=int)

        # LM properties
        self.WG_pix_size = None
        self.IM_pix_size = None

        # Mag independent properties
        self.s2ss_matrix = None

        # View properties
        self.view_pix_size = None
        self.view_c2ss_matrix = None
        self.view_ss2is_matrix = None           # Should be independent between mags, but different for LM, so safe both, view and rec, to be safe
        self.view_ta_rotation = None
        self.view_rotM = None
        self.view_beam_diameter = None

        # Rec properties
        self.rec_pix_size = None
        self.rec_c2ss_matrix = None
        self.rec_ss2is_matrix = None
        self.rec_ta_rotation = None
        self.rec_rotM = None
        self.rec_beam_diameter = None
        
        # Read and overwrite params with values from file
        if file_dir and (file_dir / self.file_name).exists():
            self.readFromFile(Path(file_dir))
            log(f"NOTE: Instantiated imaging parameters from {self.file_name} in {file_dir}!")
        else:
            log(f"NOTE: Instantiated blank imaging paramters because no {self.file_name} was found.")

    @dummy_skip
    @serialem_check
    def getViewParams(self, final_attempt=False):
        low_dose, area = sem.ReportLowDose()
        if low_dose and area == 0:
            cam_props = sem.CameraProperties()
            self.cam_dims = np.array([cam_props[0], cam_props[1]], dtype=int)

            self.view_pix_size = cam_props[4]   # [nm]
            self.s2ss_matrix = np.array(sem.StageToSpecimenMatrix(0)).reshape((2, 2))
            self.view_c2ss_matrix = np.array(sem.CameraToSpecimenMatrix(0)).reshape((2, 2))
            self.view_ss2is_matrix = np.array(sem.SpecimenToISMatrix(0)).reshape((2, 2))
            self.view_ta_rotation = 90 - np.degrees(np.arctan(self.view_c2ss_matrix[0, 1] / self.view_c2ss_matrix[0, 0]))
            self.view_rotM = np.array([[np.cos(np.radians(self.view_ta_rotation)), np.sin(np.radians(self.view_ta_rotation))], [-np.sin(np.radians(self.view_ta_rotation)), np.cos(np.radians(self.view_ta_rotation))]])
            self.view_beam_diameter = sem.ReportIlluminatedArea() * 100
        else:
            if not final_attempt:
                log(f"WARNING: Cannot get View parameters in current Low Dose state {low_dose}, {area}. Trying to switch to medium mag imaging state...")
                if isinstance(self.MM_image_state, list):
                    for state in self.MM_image_state:
                        sem.GoToImagingState(str(state))
                else:
                    sem.GoToImagingState(str(self.MM_image_state))
                sem.GoToLowDoseArea(0)
                self.getViewParams(final_attempt=True)
            else:
                log(f"ERROR: Cannot get View parameters in current Low Dose state {low_dose}, {area}. Check your imaging state settings!")
                sys.exit()


    @dummy_skip
    @serialem_check
    def getRecParams(self, final_attempt=False):
        low_dose, area = sem.ReportLowDose()
        if low_dose and area == 3:
            cam_props = sem.CameraProperties()
            self.cam_dims = np.array([cam_props[0], cam_props[1]], dtype=int)

            self.rec_pix_size = cam_props[4]    # [nm]
            self.s2ss_matrix = np.array(sem.StageToSpecimenMatrix(0)).reshape((2, 2))
            self.rec_c2ss_matrix = np.array(sem.CameraToSpecimenMatrix(0)).reshape((2, 2))
            self.rec_ss2is_matrix = np.array(sem.SpecimenToISMatrix(0)).reshape((2, 2))
            self.rec_ta_rotation = 90 - np.degrees(np.arctan(self.rec_c2ss_matrix[0, 1] / self.rec_c2ss_matrix[0, 0]))
            self.rec_rotM = np.array([[np.cos(np.radians(self.rec_ta_rotation)), np.sin(np.radians(self.rec_ta_rotation))], [-np.sin(np.radians(self.rec_ta_rotation)), np.cos(np.radians(self.rec_ta_rotation))]])
            self.rec_beam_diameter = sem.ReportIlluminatedArea() * 100

            if self.rec_beam_diameter < 0:
                log(f"ERROR: Illuminated area/beam diameter is negative! Please check your imaging state setup!")
                sys.exit()
        else:
            if not final_attempt:
                log(f"WARNING: Cannot get Record parameters in current Low Dose state {low_dose}, {area}. Trying to switch to medium mag imaging state...")
                if isinstance(self.MM_image_state, list):
                    for state in self.MM_image_state:
                        sem.GoToImagingState(str(state))
                else:
                    sem.GoToImagingState(str(self.MM_image_state))
                sem.GoToLowDoseArea(3)
                self.getRecParams(final_attempt=True)
            else:
                log(f"ERROR: Cannot get Record parameters in current Low Dose state {low_dose}, {area}. Check your imaging state settings!")
                sys.exit()

    def export(self, dir):
        params_file = Path(dir) / self.file_name
        with open(params_file, "w+") as f:
             json.dump(vars(self), f, indent=4, default=utils.convertArray)
        log(f"NOTE: Saved imaging parameters at {params_file}")
        
    def readFromFile(self, dir):
        params_file = Path(dir) / self.file_name
        with open(params_file, "r") as f:
            params = json.load(f, object_hook=utils.revertArray)
        for key, value in params.items():
            vars(self)[key] = value
        log(f"NOTE: Read imaging parameters from {params_file}.")



#!/usr/bin/env python
# ===================================================================
# ScriptName:   gui_lam
# Purpose:      User interface for training SPACEtomo segmentation models using nnU-Netv2
#               More information at http://github.com/eisfabian/SPACEtomo
# Author:       Fabian Eisenstein
# Created:      2024/04/26
# Revision:     v1.2
# Last Change:  2024/09/02: fixed lock editing after inspected, fixed save button, added askForSave on exit
#               2024/08/09: changed to pathlib, adjusted to Plot class changes
#               2042/07/25: added clipping of plot boxes to map bounds
#               2024/07/18: reworked GUI to class
#               2024/06/07: added lamella table row highlighting, added save coords in YOLO format
#               2024/06/05: introduced Menu class, added shortcuts for most buttons
#               2024/06/04: added pix size check for export, added rescaling of mrc file
#               2024/06/03: added include empty tiles option, added save json option
#               2024/05/31: added status line, fixed tiles bounds, added next button, added hide/showElements
#               2024/05/30: added rescaling, saving as png, fixed export tiles
#               2024/05/29: outsourced plot clearing
#               2024/05/23: added lamella labels on plot, added padding when exporting edge tiles, fixed plotTiles
#               2024/05/10: added reading and writing of YOLO format labels, added saving of tiles with labels, added map class
#               2024/05/02: added lamella color from config, added reset zoom, delegated plot generation, call lamella detection
#               2024/04/30: added lamella focus buttons, outline scaling, list sorting
#               2024/04/26: Copy most of SPACEtomo_tgt
# ===================================================================

import os
import sys
os.environ["__GLVND_DISALLOW_PATCHING"] = "1"           # helps to minimize Segmentation fault crashes on Linux when deleting textures
from pathlib import Path
try:
    import dearpygui.dearpygui as dpg
except:
    print("ERROR: DearPyGUI module not installed! If you cannot install it, please run the GUI from an external machine.")
    sys.exit()
from PIL import Image
Image.MAX_IMAGE_PIXELS = None
import numpy as np

import SPACEtomo.config as config
from SPACEtomo.modules.mod_wg import WGModel, Boxes
from SPACEtomo.modules.gui import gui
from SPACEtomo.modules.gui.menu import Menu
from SPACEtomo.modules.gui.plot import Plot, PlotBox
from SPACEtomo.modules.gui.map import MMap
from SPACEtomo.modules import utils
from SPACEtomo.modules.utils import log
from SPACEtomo import __version__


# Check if IMOD is available (needed for mrc file import)
if "IMOD" in os.environ["PATH"]:
    IMOD = True
else:
    IMOD = False

### FUNCTIONS ###
class LamellaBox(PlotBox):
    def __init__(self, coords, parent, color=gui.COLORS["error"], lamella_info=None) -> None:
        super().__init__(coords, parent, color)

        self.label = None
        if lamella_info is not None:
            self.addInfo(lamella_info)

    def addInfo(self, lamella_info):
        if self.label is None:
            self.label = lamella_info["label"]
        self.cat = lamella_info["cat"]
        self.prob = lamella_info["prob"]
        self.coords = (self.p1 + self.p2) / 2
        self.area = np.prod(np.abs((self.p2 - self.p1)))

class LamellaGUI:

    def mouseClick(self, sender, app_data):
        mouse_coords = np.array(dpg.get_plot_mouse_pos())

        # Check if click needs to be processed
        if not dpg.is_item_hovered(self.plot.plot) or not self.plot.withinBounds(mouse_coords) or self.inspected:
            return

        if dpg.is_mouse_button_down(dpg.mvMouseButton_Left) and (dpg.is_key_down(dpg.mvKey_LShift) or dpg.is_key_down(dpg.mvKey_RShift)) and len(self.plot.img) > 0:
            # Clip box dims to map bounds
            mouse_coords = np.array((np.clip(mouse_coords[0], 0, self.loaded_map.img.shape[0] * self.loaded_map.pix_size / 10000), np.clip(mouse_coords[1], 0, self.loaded_map.img.shape[1] * self.loaded_map.pix_size / 10000)))

            if len(self.plot.boxes) > 0:
                for b, box in enumerate(self.plot.boxes):
                    if box.within(mouse_coords, margin=5):
                        self.mouse_box = box
                        self.mouse_box.updateP1(mouse_coords)
                        self.plot.boxes.pop(b)
                        break
            if self.mouse_box is None:
                self.mouse_box = LamellaBox(mouse_coords, parent=self.plot.plot)
                self.scaleOutlines()

        elif dpg.is_mouse_button_down(dpg.mvMouseButton_Right) and len(self.plot.boxes) > 0:
            # Call Lamella menu by right clicking
            for b, box in enumerate(self.plot.boxes):
                dpg.unhighlight_table_row("tbllam", b)
                if box.within(mouse_coords):
                    # Highlight selected lamella in table
                    dpg.highlight_table_row("tbllam", b, gui.COLORS["subtle"])
                    self.selected_lamella = b
                    # Show lamella window
                    self.showLamellaInfo(sender, app_data, b)

        elif dpg.is_mouse_button_down(dpg.mvMouseButton_Left) and len(self.plot.boxes) > 0:
            # Start dragging to update lamella box
            for b, box in enumerate(self.plot.boxes):
                dpg.unhighlight_table_row("tbllam", b)
                if box.within(mouse_coords, margin=min(box.width, box.height) * 0.1):
                    # Highlight selected lamella in table
                    dpg.highlight_table_row("tbllam", b, gui.COLORS["subtle"])
                    self.selected_lamella = b
                    self.drag_start = mouse_coords
                    box.startUpdate(mouse_coords)

    def mouseDrag(self, sender, app_data):
        mouse_coords = np.array(dpg.get_plot_mouse_pos())

        # Update box boundary when drawing
        if self.mouse_box is not None:
            self.mouse_box.updateP2(mouse_coords)
        
        # Update box drag position
        elif self.drag_start is not None:
            self.plot.boxes[self.selected_lamella].update(mouse_coords)

    def mouseRelease(self, sender, app_data):
        mouse_coords = np.array(dpg.get_plot_mouse_pos())

        # End drawing of lamella box
        if self.mouse_box is not None:
            # Clip box dims to map bounds
            mouse_coords = np.array((np.clip(mouse_coords[0], 0, self.loaded_map.img.shape[0] * self.loaded_map.pix_size / 10000), np.clip(mouse_coords[1], 0, self.loaded_map.img.shape[1] * self.loaded_map.pix_size / 10000)))

            self.mouse_box.updateP2(mouse_coords)
            self.plot.boxes.append(self.mouse_box)
            self.mouse_box = None

            # Add lamella data
            if len(self.plot.boxes) > 1:
                index = int(self.plot.boxes[-2].label.split("L")[-1]) + 1
            else:
                index = 1
            lamella_info = {"label": "L" + str(index).zfill(2), "cat": config.WG_model_categories.index("good"), "prob": 1}
            self.plot.boxes[-1].addInfo(lamella_info)
            self.plot.boxes[-1].updateColor(config.WG_model_gui_colors[config.WG_model_categories.index("good")])
            self.plot.boxes[-1].drawLabel(self.plot.boxes[-1].label)
            self.plot.sortList("boxes", "label")

            # Update table
            self.makeLamellaTable()

            dpg.show_item(self.menu.all_elements["btn_save"])

        # End dragging of lamella box
        elif self.drag_start is not None:
            self.drag_start = None
            self.plot.boxes[self.selected_lamella].endUpdate()
            dpg.show_item(self.menu.all_elements["btn_save"])

    def mouseScroll(self, sender, app_data):
        """Catches scroll on plot to scale outline of drawn boxes."""

        if dpg.is_item_hovered(self.plot.plot) and len(self.plot.boxes) > 0:
            self.scaleOutlines()

    def loadMap(self, sender, app_data, user_data, tile_size=None, stitched=None, quantile=None):
        selected_files = sorted([file for file in app_data["selections"].values()])
        file_path = Path(selected_files[0])

        # Have to manually overwrite args, because buttons seems to fill all args with None
        if stitched is None:
            stitched = False
        if quantile is None:
            quantile = dpg.get_value(self.menu.all_elements["inp_quantile"])

        #if not file_path.endswith((".png", ".mrc", ".map")):
        if not file_path.suffix.lower() in [".png", ".mrc", ".map"]:
            log("ERROR: Can only read .png and .mrc files!")
            gui.showInfoBox("ERROR", "Can only read .png and .mrc files!")
            return
        
        # Update GUI
        self.menu.hide()
        self.plot.updateLabel(file_path.name + " loading...")
        self.status.update("Loading map...")

        if tile_size is None:
            tile_size = (config.WG_model_sidelen, config.WG_model_sidelen)

        # Reset plot
        self.plot.clearAll()

        pix_size_png = config.WG_model_pix_size * 10 #[A] # Will be overwritten by mrc header pixel size when loading map
        dpg.set_item_label(self.plot.x_axis, "x [µm]")
        dpg.set_item_label(self.plot.y_axis, "y [µm]")

        # Load map file
        self.loaded_map = MMap(file_path, pix_size_png, stitched, tile_size, quantile=quantile, status=self.status)

        # Determine binning
        self.loaded_map.checkBinning()

        self.status.update("Plotting map...", box=True)

        map_texture = self.loaded_map.getTexture()
        dims_plot = np.flip(np.array(self.loaded_map.img.shape) * self.loaded_map.pix_size / 10000)

        # Load map to plot
        self.plot.addImg(map_texture, [[0, dims_plot[0]], [0, dims_plot[1]]], self.loaded_map.binning, label="map")

        # Fit axes
        self.plot.resetZoom()
        self.plot.updateLabel(self.loaded_map.file.name)

        self.status.update("Loading lamellae...")

        # Load detected lamellae from file
        self.loadBboxes_yolo(fallback=self.loadBboxes_json)

        # Show lamella list
        self.selected_lamella = None
        self.makeLamellaTable()

        # Show tiles for export
        self.plotTiles()

        # Update GUI
        self.menu.unlockRows(["lamlist", "detect", "visuals", "inspect", "rescale"])
        if self.loaded_map.file.suffix.lower() in [".mrc", ".map"]:
            self.menu.unlockRows(["mrcscale"])
        dpg.set_value(self.menu.all_elements["inp_pixsize"], int(self.loaded_map.pix_size / 10))

        # Check if map was already inspected
        #if os.path.exists(os.path.splitext(self.loaded_map.file)[0] + "_inspected.txt"):
        if (self.loaded_map.file.parent / (self.loaded_map.file.stem + "_inspected.txt")).exists():
            self.markInspected()
            return

        self.menu.show()
        self.status.update()

        # Check if more files are available for load next button
        # Add file_name to user_data in case rescale reloads map in png (sender also overwrites user_data because it was manually loaded)
        btn_next = self.menu.all_elements["btn_next"]
        if dpg.get_item_user_data(btn_next) is None or sender is not None:
            dpg.configure_item(btn_next, user_data=self.loaded_map.file)
            dpg.configure_item("key_next", user_data=self.loaded_map.file)
        else:
            # If new map has same extension as old map, update user_data, but keep same if map was just converted from mrc to png
            if Path(dpg.get_item_user_data(btn_next)).suffix == self.loaded_map.file.suffix:
                dpg.configure_item(btn_next, user_data=self.loaded_map.file)
                dpg.configure_item("key_next", user_data=self.loaded_map.file)

        #file_list = sorted(glob.glob(os.path.join(os.path.dirname(dpg.get_item_user_data(btn_next)), "*" + os.path.splitext(dpg.get_item_user_data(btn_next))[-1])))
        next_file = Path(dpg.get_item_user_data(btn_next))
        file_list = sorted(next_file.parent.glob("*" + next_file.suffix))

        cur_id = file_list.index(dpg.get_item_user_data(btn_next))
        if cur_id + 1 < len(file_list):
            dpg.show_item(btn_next)


    def loadNextMap(self, sender, app_data, user_data):
        if self.loaded_map is None:
            return
        
        # Load extension from button if possible
        if user_data is not None:
            extension = Path(user_data).suffix
            map_file = user_data
        else:
            extension = self.loaded_map.file.suffix
            map_file = self.loaded_map.file

        # Make list of files with same extension
        #file_list = sorted(glob.glob(os.path.join(os.path.dirname(self.loaded_map.file), "*" + extension)))
        file_list = sorted(self.loaded_map.file.parent.glob("*" + extension))

        # Get ID of current map in list
        cur_id = file_list.index(map_file)

        # Load next map and update button
        if cur_id + 1 < len(file_list):
            self.loadMap(None, {"selections": {file_list[cur_id + 1].name: file_list[cur_id + 1]}}, None)
        if cur_id + 2 >= len(file_list):
            dpg.hide_item(self.menu.all_elements["btn_next"])

    def rescaleMap(self):
        # Update GUI
        self.menu.hide()

        # Get pix size input
        rescale_pix_size = dpg.get_value(self.menu.all_elements["inp_pixsize"]) * 10 # Angstrom/px

        # Rescale map
        log("Rescaling map...")
        old_dims = np.array(self.loaded_map.img.shape)
        self.loaded_map.rescale(rescale_pix_size, padding=(config.WG_model_sidelen, config.WG_model_sidelen))

        # Save only if any changes (rescaling or padding) were made
        if not np.all(self.loaded_map.img.shape == old_dims):
            self.status.update("Saving map...")
            # Save rescaled map
            file_path = self.loaded_map.file.parent / (self.loaded_map.file.stem.split("_wg")[0] + "_" + str(int(self.loaded_map.pix_size / 10)) + "nmpx_wg.png")
            log(f"Saving map as {file_path.name}...")
            Image.fromarray(np.uint8(self.loaded_map.img * 255)).save(file_path)

            # Load rescaled map
            self.loadMap(None, {"selections": {file_path.name: file_path}}, None, tile_size=self.loaded_map.tile_size, stitched=self.loaded_map.stitched)


    def loadBboxes_json(self, fallback=None):
        bbox_file = self.loaded_map.file.parent / (self.loaded_map.file.stem + "_boxes.json")
        if not bbox_file.exists():
            log("No detected lamellae found.")
            # Check fallback
            if fallback is not None:
                log("Trying fallback...")
                fallback()
        else:
            # Read bboxes
            bboxes = Boxes(bbox_file)

            dims_img = np.flip(self.loaded_map.img.shape)

            prev_box_num = len(self.plot.boxes)
            for b, bbox in enumerate(bboxes.boxes):        
                p1 = np.array([bbox.x_min, dims_img[1] - bbox.y_min]) * self.loaded_map.pix_size / 10000
                p2 = np.array([bbox.x_max, dims_img[1] - bbox.y_max]) * self.loaded_map.pix_size / 10000

                self.plot.boxes.append(LamellaBox(p1, parent=self.plot.plot, color=config.WG_model_gui_colors[bbox.cat]))
                self.plot.boxes[-1].updateP2(p2)

                # Add lamella data
                lamella_info = {"label": "L" + str(prev_box_num + b + 1).zfill(2), "cat": bbox.cat, "prob": bbox.prob}
                self.plot.boxes[-1].addInfo(lamella_info)
                self.plot.boxes[-1].drawLabel(self.plot.boxes[-1].label)

            log(f"{len(self.plot.boxes)} detected lamellae found.")
            self.menu.unlockRows(["export", "inspect"])

    def loadBboxes_yolo(self, fallback=None):
        # Check for label file in same folder
        label_file = self.loaded_map.file.parent / (self.loaded_map.file.stem + ".txt")
        if not label_file.exists():
            # Also check for label file in typical YOLO folder structure
            label_file = self.loaded_map.file.parent.parent / "labels" / (self.loaded_map.file.stem + ".txt")
            if not label_file.exists():
                log("No detected lamellae found.")
                # Check fallback
                if fallback is not None:
                    log("Trying fallback...")
                    fallback()
                    return

        # Read labels
        with open(label_file, "r") as f:
            lines = f.readlines()
        labels = []
        for label in lines:
            labels.append([float(val) for val in label.split()])

        dims_plot = dpg.get_item_configuration(self.plot.img[utils.findIndex(self.plot.img, "label", "map")]["plot"])["bounds_max"]
        #print(dims_plot)

        prev_box_num = len(self.plot.boxes)
        for l, label in enumerate(labels):
            p1 = np.array([label[1] - label[3] / 2, 1 - label[2] - label[4] / 2]) * dims_plot
            p2 = np.array([label[1] + label[3] / 2, 1 - label[2] + label[4] / 2]) * dims_plot
            
            self.plot.boxes.append(LamellaBox(p1, parent=self.plot.plot, color=config.WG_model_gui_colors[int(label[0])]))
            self.plot.boxes[-1].updateP2(p2)

            # Add lamella data
            lamella_info = {"label": "L" + str(prev_box_num + l + 1).zfill(2), "cat": int(label[0]), "prob": 1}
            self.plot.boxes[-1].addInfo(lamella_info)
            self.plot.boxes[-1].drawLabel(self.plot.boxes[-1].label)

        log(f"{len(self.plot.boxes)} detected lamellae found.")
        self.menu.unlockRows(["export"])
        
    def makeLamellaTable(self):
        # Delete table
        if dpg.does_item_exist("tbllam"):
            dpg.delete_item("tbllam")
        if len(self.plot.boxes) > 0:
            # Make new table
            with dpg.table(label="Lamellae", tag="tbllam", parent=self.menu.all_rows["lamlist"], height=400, scrollY=True, policy=dpg.mvTable_SizingFixedFit):
                dpg.add_table_column(label="Name")
                dpg.add_table_column(label="Class")
                dpg.add_table_column(label="")
                #dpg.add_table_column(label="Coords")
                #dpg.add_table_column(label="Area")

                for l, lamella in enumerate(self.plot.boxes):
                    with dpg.table_row():
                        with dpg.table_cell():
                            dpg.add_button(label=lamella.label, callback=self.focusLamella, user_data=lamella)
                        with dpg.table_cell():
                            dpg.add_text(default_value=config.WG_model_categories[lamella.cat] + " " + str(round(lamella.prob, 2)), color=config.WG_model_gui_colors[lamella.cat])
                        with dpg.table_cell():
                            #dpg.add_button(label="Edit", callback=self.showLamellaInfo, user_data=l)
                            dpg.add_image_button(gui.makeIconEdit(), callback=self.showLamellaInfo, user_data=l, tag=f"btn_lam_edit_{l}")
                        #with dpg.table_cell():
                        #    dpg.add_text(default_value=str(round(lamella["area"])))

                    #dpg.bind_item_handler_registry("tbllamcat" + lamella.label, "tbl_click_handler")
            
            with dpg.tooltip("tbllam", delay=0.5, hide_on_activity=True):
                dpg.add_text("List of lamellae in map.\nClick on name to zoom.")

    def plotTiles(self):
        self.plot.clearSeries()
        x_vals = np.array(self.loaded_map.tile_bounds[0]) * self.loaded_map.pix_size / 10000 #scaling
        try:
            self.plot.series.append({"label": "tiles_v", "plot": dpg.add_inf_line_series(x_vals, parent=self.plot.x_axis, show=False)})
        except AttributeError:      # Backward compatibility with dearpyguy<2.0
            log(f"WARNING: Consider updating DearPyGUI to version >=2.0!")
            self.plot.series.append({"label": "tiles_v", "plot": dpg.add_vline_series(x_vals, parent=self.plot.x_axis, show=False)})
        dpg.bind_item_theme(self.plot.series[-1]["plot"], "plot_tiletheme")
        
        y_vals = (-1 * np.array(self.loaded_map.tile_bounds[1]) + self.loaded_map.img.shape[0]) * self.loaded_map.pix_size / 10000 #scaling
        try:
            self.plot.series.append({"label": "tiles_h", "plot": dpg.add_inf_line_series(y_vals, horizontal=True, parent=self.plot.x_axis, show=False)})
        except AttributeError:      # Backward compatibility with dearpyguy<2.0
            self.plot.series.append({"label": "tiles_h", "plot": dpg.add_hline_series(y_vals, parent=self.plot.x_axis, show=False)})
        dpg.bind_item_theme(self.plot.series[-1]["plot"], "plot_tiletheme")

    def toggleTiles(self):
        if len(self.plot.series) > 0:
            for series in self.plot.series:
                if dpg.is_item_shown(series["plot"]):
                    dpg.hide_item(series["plot"])
                else:
                    dpg.show_item(series["plot"])

    def toggleAdvanced(self):
        self.menu.toggleAdvanced()

    def showLamellaInfo(self, sender, app_data, user_data, update_pos=True):
        # Focus plot on lamella when called from table
        if app_data is None:
            self.focusLamella(sender, app_data, self.plot.boxes[user_data])
        
        # Get mouse position for info window
        mouse_coords_global = dpg.get_mouse_pos(local=False)

        dpg.set_value("lam_label", self.plot.boxes[user_data].label)
        dpg.set_value("lam_cat", config.WG_model_categories[self.plot.boxes[user_data].cat])
        dpg.configure_item("lam_cat", user_data=user_data)
        dpg.configure_item("lam_btndel", user_data=user_data)
        dpg.configure_item("lam_btnup", user_data=user_data)
        dpg.configure_item("lam_btndown", user_data=user_data)
        if update_pos:
            dpg.set_item_pos("LAM", mouse_coords_global)
        dpg.show_item("LAM")

    def deleteLamella(self, sender, app_data, user_data):
        log(f"Deleted lamella: {self.plot.boxes[user_data].label} at {np.round((self.plot.boxes[user_data].p1 + self.plot.boxes[user_data].p2) / 2)}")
        self.plot.boxes[user_data].remove()
        self.plot.boxes.pop(user_data)
        # Reset lamella menu
        dpg.hide_item("LAM")
        dpg.configure_item("lam_cat", user_data=None)
        self.selected_lamella -= 1
        # Update table
        self.makeLamellaTable()
        dpg.show_item(self.menu.all_elements["btn_save"])

    def catLamella(self, sender, app_data, user_data):
        # Deal with call from shortcuts
        if user_data is None:
            user_data = dpg.get_item_user_data("lam_cat")
        if user_data is not None:
            self.plot.boxes[user_data].cat = config.WG_model_categories.index(app_data)
            self.plot.boxes[user_data].prob = 1
            self.plot.boxes[user_data].updateColor(config.WG_model_gui_colors[config.WG_model_categories.index(app_data)])
            # Reset lamella menu
            dpg.hide_item("LAM")
            dpg.configure_item("lam_cat", user_data=None)
            # Update table
            self.makeLamellaTable()
            dpg.show_item(self.menu.all_elements["btn_save"])

    def reorderLamellaUp(self, sender, app_data, user_data):
        """Moves lamella up in list and renames it."""

        if user_data is not None:
            box_id = user_data
            new_box_id = user_data - 1
            if new_box_id < 0: 
                new_box_id = len(self.plot.boxes) - 1

            self.plot.boxes[new_box_id], self.plot.boxes[box_id] = self.plot.boxes[box_id], self.plot.boxes[new_box_id]
            self.relabelLamellae()
            self.makeLamellaTable()
            self.showLamellaInfo(None, None, user_data=new_box_id, update_pos=False)
            dpg.show_item(self.menu.all_elements["btn_save"])

    def reorderLamellaDown(self, sender, app_data, user_data):
        """Moves lamella down in list and renames it."""

        if user_data is not None:
            box_id = user_data
            new_box_id = user_data + 1
            if new_box_id >= len(self.plot.boxes): 
                new_box_id = 0

            self.plot.boxes[new_box_id], self.plot.boxes[box_id] = self.plot.boxes[box_id], self.plot.boxes[new_box_id]
            self.relabelLamellae()
            self.makeLamellaTable()
            self.showLamellaInfo(None, None, user_data=new_box_id, update_pos=False)
            dpg.show_item(self.menu.all_elements["btn_save"])

    def relabelLamellae(self):
        """Relabels lamellae by order."""

        for b, box in enumerate(self.plot.boxes):
            box.label = "L" + str(b + 1).zfill(2)
            box.drawLabel(box.label)

    def saveLamellae(self, sender, app_data, user_data, map_file=None, map_pix_size=None):
        if map_file is None:
            map_file = self.loaded_map.file
        if map_pix_size is None:
            map_pix_size = self.loaded_map.pix_size
            map_y = self.loaded_map.img.shape[0]
        else:
            map_y = int(self.loaded_map.img.shape[0] * self.loaded_map.pix_size / map_pix_size)

        bboxes = []
        for plot_box in self.plot.boxes:
            bbox = np.zeros(6)
            # Convert plot coords to pixel coords
            bbox[0], bbox[1] = plot_box.p1 / map_pix_size * 10000
            bbox[1] = map_y - bbox[1]
            bbox[2], bbox[3] = plot_box.p2 / map_pix_size * 10000
            bbox[3] = map_y - bbox[3]
            # Add class
            bbox[4] = plot_box.cat
            bbox[5] = plot_box.prob

            bboxes.append(bbox)
        bboxes = np.array(bboxes)

        # Save bboxes
        box_file = map_file.parent / (map_file.stem + "_boxes.json")
        bboxes = Boxes(bboxes, pix_size=map_pix_size / 10, img_size=self.loaded_map.img.shape)
        bboxes.saveFile(box_file)

        #with open(os.path.splitext(map_file)[0].split("_wg")[0] + "_boxes.json", "w+") as f:
        #with open(box_file, "w+") as f:
        #    json.dump(bboxes, f, indent=4, default=utils.convertArray)
        
        log(f"Saved lamellae to: {box_file}")
        dpg.hide_item(self.menu.all_elements["btn_save"])

    def saveLabel(self):        # in YOLO format
        # Get bboxes
        labels = []
        for box in self.plot.boxes:
            c = box.cat
            # Scale coords to relative img coords
            x, y = box.coords * 10000 / self.loaded_map.pix_size / np.array(self.loaded_map.img.shape)
            w, h = np.abs((box.p2 - box.p1)) * 10000 / self.loaded_map.pix_size / np.array(self.loaded_map.img.shape)
            # Invert y axis
            y = 1 - y
            labels.append([c, x, y, w, h])
        # Write txt file
        label_file = self.loaded_map.file.parent / (self.loaded_map.file.stem + ".txt")
        #with open(os.path.splitext(self.loaded_map.file)[0] + ".txt", "w+") as f: 
        with open(label_file, "w+") as f: 
            for label in labels:
                f.write(" ".join(str(val) for val in label) + "\n")   
        log(f"Saved lamellae in YOLO format to: {label_file}") 

    def focusLamella(self, sender, app_data, user_data):
        plot_box = user_data
        self.selected_lamella = self.plot.boxes.index(plot_box)

        # Highlight lamella in table
        for i in range(len(self.plot.boxes)):
            dpg.unhighlight_table_row("tbllam", i)
        dpg.highlight_table_row("tbllam", self.selected_lamella, gui.COLORS["subtle"])

        # Get FOV for plot
        margin = 50

        min_x = min([plot_box.p1[0], plot_box.p2[0]]) - margin
        max_x = max([plot_box.p1[0], plot_box.p2[0]]) + margin
        min_y = min([plot_box.p1[1], plot_box.p2[1]]) - margin
        max_y = max([plot_box.p1[1], plot_box.p2[1]]) + margin

        # Get old limits to determine aspect ratio
        x_lim = dpg.get_axis_limits(self.plot.x_axis)
        y_lim = dpg.get_axis_limits(self.plot.y_axis)
        aspect = (x_lim[1] - x_lim[0]) / (y_lim[1] - y_lim[0])

        # Lock axes to center lamella
        dpg.set_axis_limits(self.plot.y_axis, min_y, max_y)
        dpg.set_axis_limits(self.plot.x_axis, min_x + margin - (max_y - min_y) * aspect / 2, min_x + margin + (max_y - min_y) * aspect / 2)

        # Unlock the axes one frame later so axis limits are applied first
        dpg.set_frame_callback(dpg.get_frame_count() + 1, callback=self.unlockAxes)

    def nextLamella(self):
        if len(self.plot.boxes) > 0:
            if self.selected_lamella is None:
                self.selected_lamella = 0
            else:
                self.selected_lamella += 1
            if self.selected_lamella >= len(self.plot.boxes):
                self.selected_lamella = 0
            box = self.plot.boxes[self.selected_lamella]
            self.focusLamella(None, None, box)

    def prevLamella(self):
        if len(self.plot.boxes) > 0:
            if self.selected_lamella is None:
                self.selected_lamella = 0
            else:
                self.selected_lamella -= 1
            if self.selected_lamella < 0:
                self.selected_lamella = len(self.plot.boxes) - 1
            box = self.plot.boxes[self.selected_lamella]
            self.focusLamella(None, None, box)

    def unlockAxes(self):
        dpg.set_axis_limits_auto(self.plot.x_axis)
        dpg.set_axis_limits_auto(self.plot.y_axis)
        # Scale box outlines to new zoom level
        self.scaleOutlines()

    # Scale outline of boxes according to zoom level
    def scaleOutlines(self, fraction=200):
        x_lim = dpg.get_axis_limits(self.plot.x_axis)
        width = x_lim[1] - x_lim[0]
        thickness = width / fraction

        for box in self.plot.boxes:
            box.updateThickness(thickness)

        if self.mouse_box is not None:
            self.mouse_box.updateThickness(thickness)

    def callLamellaDetection(self, sender, app_data, user_data):
        # Check for info box input
        if user_data is not None and dpg.does_item_exist(user_data[0]):
            dpg.delete_item(user_data[0])
            dpg.split_frame()
            # Deal with info box input
            if len(user_data) > 1 and user_data[1] == 0:
                dpg.set_value(self.menu.all_elements["inp_pixsize"], config.WG_model_pix_size)
                self.rescaleMap()        
            else:
                self.menu.show()
                return
        
        # Update GUI
        self.menu.hide()

        # Check if map has correct pixel size (within 10% of model pixel size)
        if abs(self.loaded_map.pix_size - config.WG_model_pix_size * 10) > 0.1 * config.WG_model_pix_size:
            log("WARNING: Loaded map does not have a suitable pixel size for lamella detection model. Rescaling...")
            gui.showInfoBox("WARNING", "The loaded map does not have a suitable pixel size for the lamella detection model.\nThe map will be rescaled to " + str(config.WG_model_pix_size) + " nm/px.", self.callLamellaDetection, ["OK", "Cancel"])
            return

        # Reset boxes
        #for box in plot.boxes:
        #    box.remove()
        self.plot.clearBoxes()

        log("Searching for lamellae...")
        self.status.update("Searching lamellae...", box=True)
        WG_model = WGModel()
        WG_model.findLamellae(self.loaded_map.file.parent, self.loaded_map.file.name, suffix="")
        self.loadBboxes_json()

        # Show lamella list
        self.makeLamellaTable()

        # Update GUI
        self.menu.show()
        self.status.update()
        dpg.set_value(self.menu.all_elements["inp_pixsize"], int(self.loaded_map.pix_size / 10))

    @staticmethod
    def test(sender, app_data, user_data):
        log(f"Test: {sender}, {app_data}, {user_data}")

    def splitTilesForExport(self, sender, app_data, user_data):
        # Check for info box input
        if user_data is not None and dpg.does_item_exist(user_data[0]):
            dpg.delete_item(user_data[0])
            dpg.split_frame()
            if len(user_data) > 1 and user_data[1] == 0:
                # Rescale
                dpg.set_value(self.menu.all_elements["inp_pixsize"], config.WG_model_pix_size)
                map_file = self.loaded_map.file.parent / (self.loaded_map.file.stem + "_" + str(config.WG_model_pix_size) + "nmpx_wg.png")
                self.saveLamellae(None, None, None, map_file=map_file, map_pix_size=config.WG_model_pix_size * 10)
                self.rescaleMap()
            elif len(user_data) > 1 and user_data[1] == 2:
                # Cancel
                return
        else:
            # Check if map has correct pixel size (within 10% of model pixel size)
            if abs(self.loaded_map.pix_size - config.WG_model_pix_size * 10) > 0.1 * config.WG_model_pix_size:
                log("WARNING: Loaded map does not have a suitable pixel size for lamella detection model. Rescaling?")
                gui.showInfoBox("WARNING", "The loaded map does not have a suitable pixel size for the lamella detection model.\nThe map should be rescaled to " + str(config.WG_model_pix_size) + " nm/px.", self.splitTilesForExport, ["Rescale", "Export anyways", "Cancel"])
                return
            
        # Check checkboxes
        save_empty_tiles = dpg.get_value(self.menu.all_elements["chk_empty"])
        use_model_tile_size = dpg.get_value(self.menu.all_elements["chk_modeltiles"])
        use_padding = dpg.get_value(self.menu.all_elements["chk_padding"])

        # Change map tile size and bounds
        if use_model_tile_size:
            self.loaded_map.changeTiling(np.array((config.WG_model_sidelen, config.WG_model_sidelen)))
            self.plotTiles()
        
        self.menu.hide()
        self.status.update("Exporting tiles...")
        # Get bboxes
        bboxes = []
        for box in self.plot.boxes:
            c = box.cat
            # Scale coords to img coords
            x, y = box.coords * 10000 / self.loaded_map.pix_size
            w, h = np.abs((box.p2 - box.p1)) * 10000 / self.loaded_map.pix_size# + np.array([10, 10])
            # Invert y axis
            y = self.loaded_map.img.shape[0] - y
            bboxes.append([c, x, y, w, h])

        # Go through all tiles (+ 1 for partial tiles)
        total_tile_count = 0
        total_lam_count = 0
        for i in range(self.loaded_map.tile_num[0] + 1):
            for j in range(self.loaded_map.tile_num[1] + 1):
                labels = []
                tile_img, tile_bounds = self.loaded_map.returnTile(i, j, padding=use_padding)
                if np.any(tile_img):
                    for bbox in bboxes:
                        # Convert to lower and upper bounds [[xmin, xmax], [ymin, ymax]]
                        bounds = [[bbox[2] - bbox[4] / 2, bbox[2] + bbox[4] / 2], [bbox[1] - bbox[3] / 2, bbox[1] + bbox[3] / 2]]
                        overlap, overlap_bounds = self.overlapBoxes(bounds, tile_bounds)
                        if overlap:
                            # Get bounds in coords within tile image (subtract lower tile bounds)
                            rel_bounds = [np.array(overlap_bounds[0]) - tile_bounds[0][0], np.array(overlap_bounds[1]) - tile_bounds[1][0]]
                            # Get relative coordinates for YOLO [xc, yc, w, h]
                            #yolo_bbox = [(bbox[1] - tile_bounds[1][0]) / loaded_map.tile_size[1], (bbox[2] - tile_bounds[0][0]) / loaded_map.tile_size[0], bbox[3] / loaded_map.tile_size[1], bbox[4] / loaded_map.tile_size[0]]
                            yolo_bbox = [np.sum(rel_bounds[1]) / 2 / self.loaded_map.tile_size[1], np.sum(rel_bounds[0]) / 2 / self.loaded_map.tile_size[0], np.diff(rel_bounds[1])[0] / self.loaded_map.tile_size[1], np.diff(rel_bounds[0])[0] / self.loaded_map.tile_size[0]]
                            # Add label
                            labels.append([bbox[0]] + yolo_bbox)

                    if len(labels) > 0 or save_empty_tiles:
                        self.status.update("Exporting tile [" + str(i) + ", " + str(j) + "]...")
                        #if not os.path.exists(os.path.join(os.path.dirname(self.loaded_map.file), "YOLO_dataset")): 
                        #    os.makedirs(os.path.join(os.path.dirname(self.loaded_map.file), "YOLO_dataset"))
                        if not (self.loaded_map.file.parent / "YOLO_dataset").exists(): 
                            (self.loaded_map.file.parent / "YOLO_dataset").mkdir()
                        #self.exportTile(os.path.join(os.path.dirname(self.loaded_map.file), "YOLO_dataset", os.path.splitext(os.path.basename(self.loaded_map.file))[0] + "_" + str(i) + "_" + str(j) + ".png"), tile_img, labels)
                        self.exportTile(self.loaded_map.file.parent / "YOLO_dataset" / (self.loaded_map.file.stem + f"_{i}_{j}.png"), tile_img, labels)
                        total_tile_count += 1
                        total_lam_count += len(labels)

        # Update GUI
        self.menu.show()
        self.status.update()
        dpg.set_value(self.menu.all_elements["inp_pixsize"], int(self.loaded_map.pix_size / 10))

        log(f"Successfully exported {total_tile_count} tiles containing {total_lam_count} lamellae to {self.loaded_map.file.parent / 'YOLO_dataset'}")
        gui.showInfoBox("INFO", "Successfully exported " + str(total_tile_count) + " tiles containing " + str(total_lam_count) + " lamellae!")

    @staticmethod
    def overlapBoxes(bounds1, bounds2):
        if bounds1[0][0] > bounds2[0][1] or bounds1[0][1] < bounds2[0][0] or bounds1[1][0] > bounds2[1][1] or bounds1[1][1] < bounds2[1][0]:
            return False, []
        else: 
            x_min = max(bounds1[0][0], bounds2[0][0])
            x_max = min(bounds1[0][1], bounds2[0][1])
            y_min = max(bounds1[1][0], bounds2[1][0])
            y_max = min(bounds1[1][1], bounds2[1][1])
            return True, [[x_min, x_max], [y_min, y_max]]

    @staticmethod
    def exportTile(name, img, labels):
        if img.dtype == np.uint8:
            Image.fromarray(img).save(name)
        else:
            Image.fromarray((img * 255).astype(np.uint8)).save(name)
        with open(name.parent / (name.stem + ".txt"), "w+") as f: 
            for label in labels:
                f.write(" ".join(str(val) for val in label) + "\n")
        log(f"Exported {len(labels)} lamellae: {name}")

    def datasetComposition(self):
        if self.loaded_map is not None and (self.loaded_map.file.parent / "YOLO_dataset").exists():
            dataset_dir = self.loaded_map.file.parent / "YOLO_dataset"
        else:
            dataset_dir = Path.cwd()

        labelfiles = sorted(dataset_dir.glob("**/*.txt"))

        total = 0
        categories = np.zeros(len(config.WG_model_categories))
        for file in labelfiles:
            with open(file) as f:
                lines = f.readlines()

            total += len(lines)
            for line in lines:
                categories[int(line.split()[0])] += 1

        log(f"Dataset composition in {dataset_dir}:")
        message = "Number of images: " + str(len(labelfiles)) + "\nTotal lamellae:   " + str(total) + "\n\nCategories: "
        if total > 0:
            categories = sorted(zip(categories, config.WG_model_categories), reverse=True)
            for cat in categories:
                percent = str(round(cat[0] / total * 100))
                message += "\n" + cat[1] + ":" + " " * (20 - len(cat[1]) - len(str(cat[0]))) + str(int(cat[0])) + " " * (6 - len(percent)) + "(" + percent + "%)"
        else:
            message += "None"
        log(message)
        gui.showInfoBox("Dataset composition", message)

    def markInspected(self):
        """Creates inspected.txt file and locks down editing."""

        # Save targets if unsaved changes
        if dpg.is_item_shown(self.menu.all_elements["btn_save"]):
            self.saveLamellae(None, None, None)

        (self.loaded_map.file.parent / (self.loaded_map.file.stem + "_inspected.txt")).touch()
        self.inspected = True

        self.menu.hide()
        self.status.update("\nLamellae were marked\nas inspected.\n(Editing disabled)", color=gui.COLORS["heading"])

        if self.auto_close:
            log(f"NOTE: Finished inspecting detected lamellae and closed GUI.")
            dpg.stop_dearpygui()

    def askForSave(self):
        """Opens popup if there are unsaved changes."""

        if dpg.is_item_shown(self.menu.all_elements["btn_save"]):
            gui.showInfoBox("WARNING", "There are unsaved changes to your lamellae!", callback=self.saveAndClose, options=["Save", "Discard"], options_data=[True, False])
        else:
            dpg.stop_dearpygui()

    def saveAndClose(self, sender, app_data, user_data):
        """Saves targets and closes GUI."""

        # Save targets if user clicked Save
        if user_data[1]:
            self.saveLamellae(None, None, None)

        # Change exit callback to avoid infinite loop
        dpg.set_exit_callback(dpg.stop_dearpygui)
        dpg.stop_dearpygui()

    @staticmethod
    def showHelp():
        message = "Keyboard shortcuts:\n\n"
        message += "A     Toggle advanced settings\n"
        message += "C     Show current dataset composition\n"
        message += "D     Detect lamellae\n"
        message += "E     Export tiles\n"
        message += "F     Open map file dialogue\n"
        message += "H     Show help\n"
        message += "N     Load next map\n"
        message += "R     Rescale map to pixel size\n"
        message += "S     Save lamellae as SPACE .json file\n"
        message += "T     Toggle tile boundaries on plot\n"
        message += "Y     Save lamellae as YOLO .txt file\n"
        message += "Down  Show next lamella on plot\n"
        message += "Up    Show previous lamella on plot\n"

        gui.showInfoBox("Help", message)

### END FUNCTIONS ###

    def __init__(self, file=None, auto_close=False) -> None:
        log("\n########################################\nRunning SPACEtomo Lamella Selection GUI\n########################################\n")

        if file:
            self.file = Path(file)
        else:
            self.file = None

        # Automatically close GUI after map was inspected
        self.auto_close = auto_close

        # Make logo
        self.logo_dims = gui.makeLogo()

        self.configureHandlers()
        self.configureThemes()

        # Create plot object to hold all entities
        self.plot = Plot("plot")
        
        # Initialise globals
        self.menu = None
        self.status = None
        self.loaded_map = None
        self.selected_lamella = None
        self.mouse_box = None
        self.drag_start = None

        self.inspected = False              # Inspection state of current map

    def configureHandlers(self):
        """Sets up dearpygui registries and handlers."""

        # Create file dialogues
        gui.fileNav("mapfile", self.loadMap, extensions=[".png", ".mrc", ".map"])

        dpg.set_viewport_resize_callback(callback=lambda: gui.window_size_change(self.logo_dims))

        # Create event handlers
        with dpg.handler_registry() as mouse_handler:
            dpg.add_mouse_click_handler(callback=self.mouseClick)
            dpg.add_mouse_release_handler(button=dpg.mvMouseButton_Left, callback=self.mouseRelease)
            dpg.add_mouse_drag_handler(button=dpg.mvMouseButton_Left, callback=self.mouseDrag)
            dpg.add_mouse_wheel_handler(callback=self.mouseScroll)

            # Shortcuts
            dpg.add_key_press_handler(dpg.mvKey_H, callback=self.showHelp)
            dpg.add_key_press_handler(dpg.mvKey_F, callback=lambda: dpg.show_item("mapfile"))
            dpg.add_key_press_handler(dpg.mvKey_N, callback=self.loadNextMap, tag="key_next")
            dpg.add_key_press_handler(dpg.mvKey_T, callback=self.toggleTiles)
            dpg.add_key_press_handler(dpg.mvKey_A, callback=self.toggleAdvanced)
            dpg.add_key_press_handler(dpg.mvKey_D, callback=self.callLamellaDetection)
            dpg.add_key_press_handler(dpg.mvKey_S, callback=self.saveLamellae)
            dpg.add_key_press_handler(dpg.mvKey_Y, callback=self.saveLabel)
            dpg.add_key_press_handler(dpg.mvKey_R, callback=self.rescaleMap)
            dpg.add_key_press_handler(dpg.mvKey_E, callback=self.splitTilesForExport)
            dpg.add_key_press_handler(dpg.mvKey_C, callback=self.datasetComposition)
            dpg.add_key_press_handler(dpg.mvKey_Down, callback=self.nextLamella)
            dpg.add_key_press_handler(dpg.mvKey_Up, callback=self.prevLamella)

            # Setup shortcuts for class selection (loop does not work for callback lambda assignment)
            if len(config.WG_model_categories) > 0: dpg.add_key_press_handler(dpg.mvKey_NumPad0, callback=lambda: self.catLamella(None, config.WG_model_categories[0], None))
            if len(config.WG_model_categories) > 1: dpg.add_key_press_handler(dpg.mvKey_NumPad1, callback=lambda: self.catLamella(None, config.WG_model_categories[1], None))
            if len(config.WG_model_categories) > 2: dpg.add_key_press_handler(dpg.mvKey_NumPad2, callback=lambda: self.catLamella(None, config.WG_model_categories[2], None))
            if len(config.WG_model_categories) > 3: dpg.add_key_press_handler(dpg.mvKey_NumPad3, callback=lambda: self.catLamella(None, config.WG_model_categories[3], None))
            if len(config.WG_model_categories) > 4: dpg.add_key_press_handler(dpg.mvKey_NumPad4, callback=lambda: self.catLamella(None, config.WG_model_categories[4], None))
            if len(config.WG_model_categories) > 5: dpg.add_key_press_handler(dpg.mvKey_NumPad5, callback=lambda: self.catLamella(None, config.WG_model_categories[5], None))
            if len(config.WG_model_categories) > 6: dpg.add_key_press_handler(dpg.mvKey_NumPad6, callback=lambda: self.catLamella(None, config.WG_model_categories[6], None))
            if len(config.WG_model_categories) > 7: dpg.add_key_press_handler(dpg.mvKey_NumPad7, callback=lambda: self.catLamella(None, config.WG_model_categories[7], None))
            if len(config.WG_model_categories) > 8: dpg.add_key_press_handler(dpg.mvKey_NumPad8, callback=lambda: self.catLamella(None, config.WG_model_categories[8], None))
            if len(config.WG_model_categories) > 9: dpg.add_key_press_handler(dpg.mvKey_NumPad9, callback=lambda: self.catLamella(None, config.WG_model_categories[9], None))

    @staticmethod
    def configureThemes():
        """Sets up dearpygui themes."""

        gui.configureGlobalTheme()

        with dpg.theme(tag="plot_tiletheme"):
            try:
                with dpg.theme_component(dpg.mvInfLineSeries):
                    dpg.add_theme_color(dpg.mvPlotCol_Line, gui.COLORS["heading"], category=dpg.mvThemeCat_Plots)   
            except AttributeError:      # Backward compatibility with dearpyguy<2.0
                with dpg.theme_component(dpg.mvHLineSeries):
                    dpg.add_theme_color(dpg.mvPlotCol_Line, gui.COLORS["heading"], category=dpg.mvThemeCat_Plots) 
                with dpg.theme_component(dpg.mvVLineSeries):
                    dpg.add_theme_color(dpg.mvPlotCol_Line, gui.COLORS["heading"], category=dpg.mvThemeCat_Plots) 

    def show(self):
        """Structures and launches main window of GUI."""

        # Setup window
        dpg.create_viewport(title="SPACEtomo Lamella Detection", disable_close=True)
        dpg.setup_dearpygui()

        # Create main window
        with dpg.window(label="GUI", tag="GUI", no_scrollbar=True, no_scroll_with_mouse=True):

            with dpg.table(header_row=False):
                dpg.add_table_column(init_width_or_weight=200, width_fixed=True)
                dpg.add_table_column()

                with dpg.table_row():
                    with dpg.table_cell(tag="tblleft"):
                        dpg.add_text(default_value="Load your WG map", tag="l1", color=gui.COLORS["heading"])

                        self.menu = Menu()
                        self.menu.newRow(tag="load", horizontal=True, locked=False)
                        if self.file:        # If file name is given, make load map button
                            self.menu.addButton(tag="btn_load", label="Load map", callback=lambda: self.loadMap(None, {"selections": {self.file.name: self.file}}, None))
                        else:
                            self.menu.addButton(tag="btn_load", label="Find map", callback=lambda: dpg.show_item("mapfile"))
                        self.menu.addButton(tag="btn_next", label="Next", callback=self.loadNextMap, show=False)

                        self.menu.newRow(tag="lamlist")

                        self.menu.newRow(tag="detect", horizontal=True, advanced=True)
                        self.menu.addButton(tag="btn_detect", label="Detect lamellae", callback=self.callLamellaDetection)

                        self.menu.newRow(tag="visuals", horizontal=True)
                        self.menu.addButton(tag="btn_rzoom", label="Reset zoom", callback=self.plot.resetZoom)
                        self.menu.addButton(tag="btn_ttiles", label="Show tiles", callback=self.toggleTiles, advanced=True)

                        self.menu.newRow(tag="rescale", horizontal=True, advanced=True)
                        self.menu.addInput("inp_pixsize", label="[nm/px]", value=100)
                        self.menu.addButton(tag="btn_rescale", label="Rescale", callback=self.rescaleMap)

                        self.menu.newRow(tag="inspect")
                        self.menu.addButton(tag="btn_save", label="Save", callback=self.saveLamellae, show=False)
                        self.menu.addButton(tag="btn_inspect", label="Confirm inspection", callback=self.markInspected, theme="large_btn_theme")

                        self.menu.newRow(tag="export")
                        self.menu.addCheckbox(tag="chk_empty", label="Include empty tiles", value=False, advanced=True)
                        self.menu.addCheckbox(tag="chk_modeltiles", label="Use model tile size", value=True, advanced=True)
                        self.menu.addCheckbox(tag="chk_padding", label="Use padding", value=True, advanced=True)
                        self.menu.addButton(tag="btn_exptiles", label="Export tiles", callback=self.splitTilesForExport)
                        self.menu.addButton(tag="btn_datacomp", label="Dataset composition", callback=self.datasetComposition, advanced=True)

                        self.menu.newRow(tag="mrcscale", horizontal=True, advanced=True)
                        self.menu.addInput(tag="inp_quantile", label="quantile", value=0.01)
                        self.menu.addButton(tag="btn_rescalemrc", label="Rescale mrc", callback=lambda: self.loadMap(None, {"selections": {self.loaded_map.file.name: self.loaded_map.file}}, None, quantile=dpg.get_value(self.menu.all_elements["inp_quantile"])))

                        self.status = gui.StatusLine()

                    with dpg.table_cell(tag="tblplot"):
                        self.plot.makePlot(x_axis_label="x [µm]", y_axis_label="y [µm]", width=-1, height=-1, equal_aspects=True, no_menus=True, crosshairs=True, pan_button=dpg.mvMouseButton_Right, no_box_select=True)

            # Create tooltips
            with dpg.tooltip("l1", delay=0.5):
                dpg.add_text("Select an .mrc or .png whole grid montage.\nIf IMOD is available, the piece coordinates will be read from the mrc header.")
            with dpg.tooltip(self.menu.all_elements["btn_detect"], delay=0.5):
                dpg.add_text("Use SPACEtomo detection model to find lamellae.")
            with dpg.tooltip(self.menu.all_elements["btn_save"], delay=0.5):
                dpg.add_text("Save lamellae coordinates for SPACEtomo.")
            with dpg.tooltip(self.menu.all_elements["btn_rescale"], delay=0.5):
                dpg.add_text("Rescale mrc map to desired pixel size. If a png file is loaded, the pixel size of the SPACEtomo model is assumed.")
            with dpg.tooltip(self.menu.all_elements["btn_exptiles"], delay=0.5):
                dpg.add_text("Export map tiles and lamellae boxes for YOLO training.")

            with dpg.tooltip("tblplot", delay=0.5, hide_on_activity=True):
                #dpg.add_text(default_value="", color=gui.COLORS["heading"], tag="tt_heading")
                dpg.add_text(default_value="- Shift + left click + drag to add lamella box\n- Right click to edit lamella", tag="tt_text")

            # Show logo
            dpg.add_image("logo", pos=(10, dpg.get_viewport_client_height() - 40 - self.logo_dims[0]), tag="logo_img")
            dpg.add_text(default_value="SPACEtomo", pos=(10 + self.logo_dims[1] / 2 - (30), dpg.get_viewport_client_height() - 40 - self.logo_dims[0] / 2), tag="logo_text")
            dpg.add_text(default_value="v" + __version__, pos=(10 + self.logo_dims[1] / 2 - (30), dpg.get_viewport_client_height() - 27 - self.logo_dims[0] / 2), tag="version_text")

        # Create lamella menu
        with dpg.window(label="Lamella", tag="LAM", no_scrollbar=True, no_scroll_with_mouse=True, popup=True, show=False):
            dpg.add_text(default_value="Lamella", color=gui.COLORS["heading"], tag="lam_label")
            dpg.add_radio_button(config.WG_model_categories, callback=self.catLamella, user_data=None, tag="lam_cat")
            dpg.add_button(label="Delete", callback=self.deleteLamella, user_data=None, tag="lam_btndel")
            with dpg.group(horizontal=True):
                dpg.add_button(arrow=True, direction=dpg.mvDir_Up, callback=self.reorderLamellaUp, user_data=None, tag="lam_btnup")
                dpg.add_button(arrow=True, direction=dpg.mvDir_Down, callback=self.reorderLamellaDown, user_data=None, tag="lam_btndown")

                dpg.bind_item_theme("lam_btnup", "small_btn_theme")
                dpg.bind_item_theme("lam_btndown", "small_btn_theme")
                with dpg.tooltip("lam_btnup", delay=0.5):
                    dpg.add_text("Move lamella up in list order.")
                with dpg.tooltip("lam_btndown", delay=0.5):
                    dpg.add_text("Move lamella down in list order.")

        dpg.bind_theme("global_theme")

        dpg.set_exit_callback(self.askForSave)

        dpg.set_primary_window("GUI", True)
        dpg.show_viewport()

        dpg.start_dearpygui()           # Can be replaced with render loop
